# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from datetime import datetime
from datetime import timezone
from typing import Any
from typing import Dict
from typing import Union

from neuro_san.interfaces.coded_tool import CodedTool


class DateTime(CodedTool):
    """
    CodedTool implementation which provides a current date and time
    """

    def invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> str:
        """
        :param args: An argument dictionary whose keys are the parameters
                to the coded tool and whose values are the values passed for them
                by the calling agent.  This dictionary is to be treated as read-only.

                The argument dictionary expects the following keys:
                    None

        :param sly_data: A dictionary whose keys are defined by the agent hierarchy,
                but whose values are meant to be kept out of the chat stream.

                This dictionary is largely to be treated as read-only.
                It is possible to add key/value pairs to this dict that do not
                yet exist as a bulletin board, as long as the responsibility
                for which coded_tool publishes new entries is well understood
                by the agent chain implementation and the coded_tool implementation
                adding the data is not invoke()-ed more than once.

                Keys expected for this implementation are:
                    None

        :return:
            In case of successful execution:
                The URL to the app as a string.
            otherwise:
                a text string an error message in the format:
                "Error: <error message>"
        """
        # Get current UTC time
        now = datetime.now(timezone.utc)

        # Format it in a user-friendly way
        friendly_time = now.strftime("%A, %B %d, %Y at %I:%M %p %Z")

        print("Current UTC date and time:", friendly_time)

        return str(now)

    async def async_invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> Union[Dict[str, Any], str]:
        raise NotImplementedError



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

import logging
import time
import threading
from typing import Dict

from neuro_san.internals.tool_factories.service_tool_factory_provider import ServiceToolFactoryProvider
from neuro_san.service.registries_watcher.periodic_updater.registry_observer import RegistryObserver
from neuro_san.internals.graph.registry.agent_tool_registry import AgentToolRegistry
from neuro_san.internals.graph.persistence.registry_manifest_restorer import RegistryManifestRestorer


class ManifestPeriodicUpdater:
    """
    Class implementing periodic manifest directory updates
    by watching agent files and manifest file itself.
    """

    def __init__(self, manifest_path: str, update_period_seconds: int):
        """
        Constructor.
        :param manifest_path: file path to server manifest file
        :param update_period_seconds: update period in seconds
        """
        self.manifest_path: str = manifest_path
        self.update_period_seconds: int = update_period_seconds
        self.logger = logging.getLogger(self.__class__.__name__)
        self.updater = threading.Thread(target=self._run, daemon=True)
        self.observer: RegistryObserver = RegistryObserver(self.manifest_path)
        self.tool_factory: ServiceToolFactoryProvider = \
            ServiceToolFactoryProvider.get_instance()
        self.go_run: bool = True

    def _run(self):
        """
        Function runs manifest file update cycle.
        """
        if self.update_period_seconds <= 0:
            # We should not run at all.
            return
        while self.go_run:
            time.sleep(self.update_period_seconds)
            # Check events that may have been triggered in target registry:
            modified, added, deleted = self.observer.reset_event_counters()
            if modified == added == deleted == 0:
                # Nothing happened - go on observing
                continue
            # Some events were triggered - reload manifest file
            self.logger.info("Observed events: modified %d, added %d, deleted %d",
                             modified, added, deleted)
            self.logger.info("Updating manifest file: %s", self.manifest_path)
            registries: Dict[str, AgentToolRegistry] = \
                RegistryManifestRestorer().restore(self.manifest_path)
            self.tool_factory.setup_tool_registries(registries)

    def start(self):
        """
        Start running periodic manifest updater.
        """
        self.logger.info("Starting manifest updater for %s with %d seconds period",
                         self.manifest_path, self.update_period_seconds)
        self.observer.start()
        self.updater.start()




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import AsyncIterator
from typing import Dict

from asyncio.queues import Queue

from neuro_san.internals.interfaces.async_hopper import AsyncHopper


class AsyncCollatingQueue(AsyncIterator, AsyncHopper):
    """
    AsyncIterator instance to asynchronously iterate over/consume the contents of
    a Queue as they come in.
    """
    # Constant for the end key
    END_KEY: str = "end"

    # Constant for the end message to be put in a Queue when all the messages are done
    END_MESSAGE: Dict[str, Any] = {END_KEY: True}

    def __init__(self, queue: Queue = None):
        """
        Constructor

        :param queue: The queue we will be iterating over.
                      Default value is None, indicating a standard Queue will be used.
        """
        self.queue: Queue = queue
        if self.queue is None:
            self.queue = Queue()

    def get_queue(self) -> Queue:
        """
        :return: The Queue associated with this instance
        """
        return self.queue

    def __aiter__(self) -> AsyncIterator[Dict[str, Any]]:
        """
        Self-identify as an AsyncIterator when called upon by
        the Python async framework.
        """
        return self

    async def __anext__(self) -> Dict[str, Any]:
        """
        :return: Blocks waiting to return the next item on the queue.
                Will throw StopAsyncIteration when the final item is detected
                via the is_final_item() method..
        """
        message = await self.queue.get()
        if self.is_final_item(message):
            raise StopAsyncIteration

        return message

    async def put(self, item: Any):
        """
        Fulfills AsyncHopper interface

        :param item: The item to put on the queue.
        """
        await self.queue.put(item)

    async def put_final_item(self):
        """
        Puts the final item on the queue indicating that no more data will
        be on the queue and the consumer's iteration can cease when it sees
        this item.
        """
        await self.put(self.END_MESSAGE)

    def is_final_item(self, item: Any) -> bool:
        """
        :param item: An item that has just been pulled off the queue
        :return: True if this item is considered the marker for the
                 end of data. False otherwise.
        """
        return isinstance(item, Dict) and item.get(self.END_KEY) is not None




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

import os

from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.language_models.base import BaseLanguageModel


class LangChainLlmFactory:
    """
    Interface for Factory classes creating LLM BaseLanguageModels

    Most methods take a config dictionary which consists of the following keys:

        "model_name"                The name of the model.
                                    Default if not specified is "gpt-3.5-turbo"

        "temperature"               A float "temperature" value with which to
                                    initialize the chat model.  In general,
                                    higher temperatures yield more random results.
                                    Default if not specified is 0.7

        "prompt_token_fraction"     The fraction of total tokens (not necessarily words
                                    or letters) to use for a prompt. Each model_name
                                    has a documented number of max_tokens it can handle
                                    which is a total count of message + response tokens
                                    which goes into the calculation involved in
                                    get_max_prompt_tokens().
                                    By default the value is 0.5.

        "max_tokens"                The maximum number of tokens to use in
                                    get_max_prompt_tokens(). By default this comes from
                                    the model description in this class.
    """

    def create_base_chat_model(self, config: Dict[str, Any],
                               callbacks: List[BaseCallbackHandler] = None) -> BaseLanguageModel:
        """
        Create a BaseLanguageModel from the fully-specified llm config.
        :param config: The fully specified llm config which is a product of
                    _create_full_llm_config() above.
        :param callbacks: A list of BaseCallbackHandlers to add to the chat model.
        :return: A BaseLanguageModel (can be Chat or LLM)
                Can raise a ValueError if the config's class or model_name value is
                unknown to this method.
        """
        raise NotImplementedError

    def get_value_or_env(self, config: Dict[str, Any], key: str, env_key: str) -> Any:
        """
        :param config: The config dictionary to search
        :param key: The key for the config to look for
        :param env_key: The os.environ key whose value should be gotten if either
                        the key does not exist or the value for the key is None
        """
        value = None
        if config is not None:
            value = config.get(key)

        if value is None:
            value = os.getenv(env_key)

        return value




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT


class ErrorFormatter:
    """
    Interface describing how error formatting can be called.

    Errors can come from various levels of the system - either as a result
    of the agents, or the agent framework itself.  In all cases, it benefits
    consumers of the agents' output to have consistent formatting of error
    messages regardless of source.
    """

    def format(self, agent_name: str, message: str, details: str = None) -> str:
        """
        Format an error message

        :param agent_name: A string describing the name of the agent experiencing
                the error.
        :param message: The specific message describing the error occurrence.
        :param details: An optional string describing further details of how/where the
                error occurred.  Think: traceback.
        :return: String encapsulation and/or filter of the error information
                presented in the arguments.
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any, Dict
from tornado.web import Application


class HttpServerApp(Application):
    """
    Class provides customized Tornado application for neuro-san service -
    with redefined internal logger so we can include custom request metadata.
    """
    def __init__(self, handlers):
        """
        Constructor:
        :param handlers: list of request handlers
        """
        # Call the base constructor
        super().__init__(handlers=handlers)

    def log_request(self, handler):
        request = handler.request
        metadata: Dict[str, Any] = handler.get_metadata()
        status = handler.get_status()
        duration = 1000 * request.request_time()  # in milliseconds
        # handler.logger is our custom HttpLogger
        handler.logger.info(metadata, "%d %s %s (%s) %.2fms",
                            status, request.method, request.uri, request.remote_ip, duration)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from copy import deepcopy

from neuro_san.internals.graph.registry.abstract_common_defs_config_filter \
    import AbstractCommonDefsConfigFilter


class DictionaryCommonDefsConfigFilter(AbstractCommonDefsConfigFilter):
    """
    An AbstractCommonDefsConfigFilter implementation that takes a agent tool
    registry config that may or may not contain commondefs definitions
    for dictionaries to substitute in by key.

    For example: Say in the config there is a top-level definition:

        "commondefs": {
            "replacement_values": {
                "foo": {
                    "bar": 1
                }
            }
        }

    This ConfigFilter implementation will replace any value of the string value "foo"
    with the full dictionary { "bar": 1 }.
    """

    def __init__(self, replacements: Dict[str, Any] = None):
        """
        Constructor

        :param replacements: An initial replacements dictionary to start out with
                whose (copied) contents will be updated with commondefs definitions
                in the basis_config during the filter_config() entry point.
                Default is None, indicating everything needed comes from the config.
        """
        super().__init__("replacement_values", replacements)

    def make_replacements(self, source_value: Any, replacements: Dict[str, Any]) -> Any:
        """
        Make replacements per the keys and values in the replacements dictionary

        :param source_value: The value to potentially do replacements on
        :param replacements: A dictionary of string keys to their replacements
        :return: A potentially new value if some key in the replacements dictionary
                is found to trigger a replacement, otherwise, the same source_value
                that came in.
        """

        if not isinstance(source_value, str):
            # We only modify strings in this implementation.
            # Let everything else go through unadulterated.
            return source_value

        # If the value is a string, we might have a candidate for replacement
        replacement_value: Any = replacements.get(source_value)
        if replacement_value is None:

            # Nothing in the dicitonary of replacements.
            # Leave the string as-is.
            # Move along. Nothing to see here.
            return source_value

        # Make a deepcopy in case the replacement is something
        # that gets modified independently by other filtering steps.
        replacement_value = deepcopy(replacement_value)

        return replacement_value




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

import json
import uuid

from leaf_common.parsers.field_extractor import FieldExtractor

from neuro_san.internals.graph.tools.argument_assigner import ArgumentAssigner
from neuro_san.internals.graph.tools.calling_tool import CallingTool
from neuro_san.internals.messages.message_utils import generate_response
from neuro_san.internals.run_context.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.run_context.interfaces.callable_tool import CallableTool
from neuro_san.internals.run_context.interfaces.run import Run
from neuro_san.internals.run_context.interfaces.run_context import RunContext


class BranchTool(CallingTool, CallableTool):
    """
    A CallingTool subclass which can also be a CallableTool.
    Thus, instances are able to be branch nodes in the tool call graph.
    Leaf nodes in the call graph are also these guys, they just happen to
    not call anyone else.
    """

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(self, parent_run_context: RunContext,
                 factory: AgentToolFactory,
                 arguments: Dict[str, Any],
                 agent_tool_spec: Dict[str, Any],
                 sly_data: Dict[str, Any]):
        """
        Constructor

        :param parent_run_context: The parent RunContext (if any) to pass
                             down its resources to a new RunContext created by
                             this call.
        :param factory: The AgentToolFactory used to create tools
        :param arguments: A dictionary of the tool function arguments passed in
        :param agent_tool_spec: The dictionary describing the JSON agent tool
                            to be used by the instance
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
        """
        super().__init__(parent_run_context, factory, agent_tool_spec, sly_data)
        self.arguments: Dict[str, Any] = arguments

    def get_decision_name(self) -> str:
        """
        :return: The string name of the decision the user is trying to make
        """
        decision_name = self.arguments.get("name")
        if decision_name is None:
            decision_name = str(uuid.uuid4())
        return decision_name

    def get_assignments(self) -> str:
        """
        :return: The string prompt for assigning values to the arguments to the agent.
        """
        # Get the properties of the function
        extractor: FieldExtractor = FieldExtractor()
        empty: Dict[str, Any] = {}

        agent_spec = self.get_agent_tool_spec()

        # Properties describe the function arguments
        properties: Dict[str, Any] = extractor.get_field(agent_spec, "function.parameters.properties", empty)

        assigner = ArgumentAssigner(properties)
        assignments: List[str] = assigner.assign(self.arguments)

        # Start to build a single assignments string, with one sentence for each property
        # listed (exception for name and description).
        assignments_str: str = "\n".join(assignments)
        return assignments_str

    def get_command(self) -> str:
        """
        :return: A string describing the objective of the component.
        """
        agent_spec = self.get_agent_tool_spec()
        return agent_spec.get("command", "Perform your instructions to the best of your ability.")

    async def integrate_callable_response(self, run: Run, messages: List[Any]) -> List[Any]:
        """
        :param run: The Run for the prescriptor (if any)
        :param messages: A current list of messages for the component.
        :return: An updated list of messages after this operation is done.
                This default implementation is just a pass-through of the messages argument.
        """
        new_messages: List[Any] = messages

        callable_tool_name = self.get_callable_tool_names(self.agent_tool_spec)
        if callable_tool_name is None:
            # If there is no callable_tool_name, then there is no action from the
            # callable class to integrate
            return new_messages

        while run.requires_action():
            # The tool we just called requires more information
            new_run = await self.make_tool_function_calls(run)
            new_run = await self.run_context.wait_on_run(new_run, self.journal)
            new_messages = await self.run_context.get_response()

        return new_messages

    async def build(self) -> List[Any]:
        """
        Main entry point to the class.
        :return: A List of messages produced during this process.
        """

        assignments = self.get_assignments()
        instructions = self.get_instructions()

        decision_name = self.get_decision_name()
        component_name = self.get_name()
        assistant_name = f"{decision_name}_{component_name}"
        await self.create_resources(assistant_name, instructions, assignments)

        command = self.get_command()
        run: Run = await self.run_context.submit_message(command)
        run = await self.run_context.wait_on_run(run, self.journal)

        messages = await self.run_context.get_response()

        messages = await self.integrate_callable_response(run, messages)

        response = generate_response(messages)
        return response

    def get_origin(self) -> List[Dict[str, Any]]:
        """
        :return: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        return self.run_context.get_origin()

    async def use_tool(self, tool_name: str, tool_args: Dict[str, Any], sly_data: Dict[str, Any]) -> str:
        """
        Experimental method to call a tool more directly from a subclass.

        NOTE: This method is not correctly reporting history or anything like that just yet.

        :param tool_name: The name of the tool to invoke
        :param tool_args: A dictionary of arguments to send to the tool.
        :param sly_data: private data dictionary to send to the tool.
        :return: A string representing the last received content text of the last message.
        """

        # Use the tool
        our_agent_spec = self.get_agent_tool_spec()
        callable_tool: CallableTool = self.factory.create_agent_tool(self.run_context,
                                                                     our_agent_spec,
                                                                     tool_name,
                                                                     sly_data,
                                                                     tool_args)
        message: str = await callable_tool.build()

        # We got a list of messages back as a string. Take the last.
        message_list: List[Dict[str, Any]] = json.loads(message)
        message_dict: Dict[str, Any] = message_list[-1]
        content: str = message_dict.get("content")

        return content




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

from langchain_anthropic.chat_models import ChatAnthropic
from langchain_ollama import ChatOllama
from langchain_nvidia_ai_endpoints import ChatNVIDIA
from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.language_models.base import BaseLanguageModel
from langchain_openai.chat_models.azure import AzureChatOpenAI
from langchain_openai.chat_models.base import ChatOpenAI

from neuro_san.internals.run_context.langchain.langchain_llm_factory import LangChainLlmFactory


class StandardLangChainLlmFactory(LangChainLlmFactory):
    """
    Factory class for LLM operations

    Most methods take a config dictionary which consists of the following keys:

        "model_name"                The name of the model.
                                    Default if not specified is "gpt-3.5-turbo"

        "temperature"               A float "temperature" value with which to
                                    initialize the chat model.  In general,
                                    higher temperatures yield more random results.
                                    Default if not specified is 0.7

        "prompt_token_fraction"     The fraction of total tokens (not necessarily words
                                    or letters) to use for a prompt. Each model_name
                                    has a documented number of max_tokens it can handle
                                    which is a total count of message + response tokens
                                    which goes into the calculation involved in
                                    get_max_prompt_tokens().
                                    By default the value is 0.5.

        "max_tokens"                The maximum number of tokens to use in
                                    get_max_prompt_tokens(). By default this comes from
                                    the model description in this class.
    """

    def create_base_chat_model(self, config: Dict[str, Any],
                               callbacks: List[BaseCallbackHandler] = None) -> BaseLanguageModel:
        """
        Create a BaseLanguageModel from the fully-specified llm config.
        :param config: The fully specified llm config which is a product of
                    _create_full_llm_config() above.
        :param callbacks: A list of BaseCallbackHandlers to add to the chat model.
        :return: A BaseLanguageModel (can be Chat or LLM)
                Can raise a ValueError if the config's class or model_name value is
                unknown to this method.
        """
        # Construct the LLM
        llm: BaseLanguageModel = None
        chat_class: str = config.get("class")
        if chat_class is not None:
            chat_class = chat_class.lower()

        model_name: str = config.get("model_name")

        if chat_class == "openai":
            llm = ChatOpenAI(
                            model_name=model_name,
                            temperature=config.get("temperature"),
                            openai_api_key=self.get_value_or_env(config, "openai_api_key",
                                                                 "OPENAI_API_KEY"),
                            openai_api_base=self.get_value_or_env(config, "openai_api_base",
                                                                  "OPENAI_API_BASE"),
                            openai_organization=self.get_value_or_env(config, "openai_organization",
                                                                      "OPENAI_ORG_ID"),
                            openai_proxy=self.get_value_or_env(config, "openai_organization",
                                                               "OPENAI_PROXY"),
                            request_timeout=config.get("request_timeout"),
                            max_retries=config.get("max_retries"),
                            presence_penalty=config.get("presence_penalty"),
                            frequency_penalty=config.get("frequency_penalty"),
                            seed=config.get("seed"),
                            logprobs=config.get("logprobs"),
                            top_logprobs=config.get("top_logprobs"),
                            logit_bias=config.get("logit_bias"),
                            streaming=True,     # streaming is always on. Without it token counting will not work.
                            n=1,                # n is always 1.  neuro-san will only ever consider one chat completion.
                            top_p=config.get("top_p"),
                            max_tokens=config.get("max_tokens"),    # This is always for output
                            tiktoken_model_name=config.get("tiktoken_model_name"),
                            stop=config.get("stop"),

                            # Set stream_usage to True in order to get token counting chunks.
                            stream_usage=True,
                            callbacks=callbacks)
        elif chat_class == "azure-openai":
            model_kwargs: Dict[str, Any] = {
                "stream_options": {
                    "include_usage": True
                }
            }
            llm = AzureChatOpenAI(
                            model_name=model_name,
                            temperature=config.get("temperature"),
                            openai_api_key=self.get_value_or_env(config, "openai_api_key",
                                                                 "OPENAI_API_KEY"),
                            openai_api_base=self.get_value_or_env(config, "openai_api_base",
                                                                  "OPENAI_API_BASE"),
                            openai_organization=self.get_value_or_env(config, "openai_organization",
                                                                      "OPENAI_ORG_ID"),
                            openai_proxy=self.get_value_or_env(config, "openai_organization",
                                                               "OPENAI_PROXY"),
                            request_timeout=config.get("request_timeout"),
                            max_retries=config.get("max_retries"),
                            presence_penalty=config.get("presence_penalty"),
                            frequency_penalty=config.get("frequency_penalty"),
                            seed=config.get("seed"),
                            logprobs=config.get("logprobs"),
                            top_logprobs=config.get("top_logprobs"),
                            logit_bias=config.get("logit_bias"),
                            streaming=True,     # streaming is always on. Without it token counting will not work.
                            n=1,                # n is always 1.  neuro-san will only ever consider one chat completion.
                            top_p=config.get("top_p"),
                            max_tokens=config.get("max_tokens"),    # This is always for output
                            tiktoken_model_name=config.get("tiktoken_model_name"),
                            stop=config.get("stop"),

                            # Azure-specific
                            azure_endpoint=self.get_value_or_env(config, "azure_endpoint",
                                                                 "AZURE_OPENAI_ENDPOINT"),
                            deployment_name=config.get("deployment_name"),
                            openai_api_version=self.get_value_or_env(config, "openai_api_version",
                                                                     "OPENAI_API_VERSION"),

                            # AD here means "ActiveDirectory"
                            azure_ad_token=self.get_value_or_env(config, "azure_ad_token",
                                                                 "AZURE_OPENAI_AD_TOKEN"),
                            model_version=config.get("model_version"),
                            openai_api_type=self.get_value_or_env(config, "openai_api_type",
                                                                  "OPENAI_API_TYPE"),
                            # Needed for token counting
                            model_kwargs=model_kwargs,
                            callbacks=callbacks)
        elif chat_class == "anthropic":
            llm = ChatAnthropic(
                            model_name=model_name,
                            max_tokens=config.get("max_tokens"),    # This is always for output
                            temperature=config.get("temperature"),
                            top_k=config.get("top_k"),
                            top_p=config.get("top_p"),
                            default_request_timeout=config.get("default_request_timeout"),
                            max_retries=config.get("max_retries"),
                            stop_sequences=config.get("stop_sequences"),
                            anthropic_api_url=self.get_value_or_env(config, "anthropic_api_url",
                                                                    "ANTHROPIC_API_URL"),
                            anthropic_api_key=self.get_value_or_env(config, "anthropic_api_key",
                                                                    "ANTHROPIC_API_KEY"),
                            streaming=True,     # streaming is always on. Without it token counting will not work.
                            # Set stream_usage to True in order to get token counting chunks.
                            stream_usage=True,
                            callbacks=callbacks)
        elif chat_class == "ollama":
            # Higher temperature is more random
            llm = ChatOllama(
                            model=model_name,
                            mirostat=config.get("mirostat"),
                            mirostat_eta=config.get("mirostat_eta"),
                            mirostat_tau=config.get("mirostat_tau"),
                            num_ctx=config.get("num_ctx"),
                            num_gpu=config.get("num_gpu"),
                            num_thread=config.get("num_thread"),
                            num_predict=config.get("num_predict", config.get("max_tokens")),
                            repeat_last_n=config.get("repeat_last_n"),
                            repeat_penalty=config.get("repeat_penalty"),
                            temperature=config.get("temperature"),
                            seed=config.get("seed"),
                            stop=config.get("stop"),
                            tfs_z=config.get("tfs_z"),
                            top_k=config.get("top_k"),
                            top_p=config.get("top_p"),
                            keep_alive=config.get("keep_alive"),
                            base_url=config.get("base_url"),

                            callbacks=callbacks)
        elif chat_class == "nvidia":
            # Higher temperature is more random
            llm = ChatNVIDIA(
                            base_url=config.get("base_url"),
                            model=model_name,
                            temperature=config.get("temperature"),
                            max_tokens=config.get("max_tokens"),
                            top_p=config.get("top_p"),
                            seed=config.get("seed"),
                            stop=config.get("stop"),
                            nvidia_api_key=self.get_value_or_env(config, "nvidia_api_key",
                                                                 "NVIDIA_API_KEY"),
                            nvidia_base_url=self.get_value_or_env(config, "nvidia_base_url",
                                                                  "NVIDIA_BASE_URL"),
                            callbacks=callbacks)
        elif chat_class is None:
            raise ValueError(f"Class name {chat_class} for model_name {model_name} is unspecified.")
        else:
            raise ValueError(f"Class {chat_class} for model_name {model_name} is unrecognized.")

        return llm



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import List
from typing import Dict

from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.message_processing.basic_message_processor import BasicMessageProcessor
from neuro_san.test.interfaces.agent_evaluator import AgentEvaluator
from neuro_san.test.interfaces.assert_forwarder import AssertForwarder


class AbstractAgentEvaluator(AgentEvaluator):
    """
    Abstract AgentEvaluator implementation that does common preparation for tests
    """

    def __init__(self, asserts: AssertForwarder, negate: bool = False):
        """
        Constructor

        :param asserts: The AssertForwarder instance to handle failures
        :param negate: If true the verification criteria should be negated
        """
        self.asserts: AssertForwarder = asserts
        self.negate: bool = negate

    def evaluate(self, processor: BasicMessageProcessor, test_key: str, verify_for: Any):
        """
        Evaluate the contents of the BasicMessageProcessor

        :param processor: The BasicMessageProcessor to evaluate
        :param test_key: the compound .-delimited key of the response value to test
        :param verify_for: The data to evaluate the response against
        """

        split: str = test_key.split(".")
        first_component: str = split[0]

        # Make a dictionary out of what we want to test
        test_dict: Dict[str, Any] = {}
        if first_component == "text":
            test_dict[first_component] = processor.get_answer()
        elif first_component == "sly_data":
            test_dict[first_component] = processor.get_sly_data()

        # Get the value we want to test out of that dictionary
        extractor = DictionaryExtractor(test_dict)
        test_value: Any = extractor.get(test_key)
        self.asserts.assertIsNotNone(test_value, f"{test_key} is None")

        # Prepare a list of keywords to verify
        verify_all: List[Any] = verify_for
        if not isinstance(verify_for, List):
            verify_all = [verify_for]
        self.asserts.assertIsInstance(verify_all, List)

        # Be sure everything specified to test for is in there... or not
        for verify_one in verify_all:
            self.test_one(verify_one, test_value)

    def test_one(self, verify_value: Any, test_value: Any):
        """
        Subclasses override this with the details of how they are testing
        the test_value from the test instance against the constancy of the verify_value.

        :param verify_value: The value to verify against
        :param test_value: The value appearing in the test sample
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: neuro_san/api/grpc/agent.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from neuro_san.api.grpc import chat_pb2 as neuro__san_dot_api_dot_grpc_dot_chat__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1eneuro_san/api/grpc/agent.proto\x12)dev.cognizant_ai.neuro_san.api.grpc.agent\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1cgoogle/api/annotations.proto\x1a\x1dneuro_san/api/grpc/chat.proto\"\x11\n\x0f\x46unctionRequest\"e\n\x08\x46unction\x12 \n\x0b\x64\x65scription\x18\x01 \x01(\tR\x0b\x64\x65scription\x12\x37\n\nparameters\x18\x02 \x01(\x0b\x32\x17.google.protobuf.StructR\nparameters\"c\n\x10\x46unctionResponse\x12O\n\x08\x66unction\x18\x01 \x01(\x0b\x32\x33.dev.cognizant_ai.neuro_san.api.grpc.agent.FunctionR\x08\x66unction\"s\n\nChatFilter\x12\x65\n\x10\x63hat_filter_type\x18\x01 \x01(\x0e\x32\x39.dev.cognizant_ai.neuro_san.api.grpc.agent.ChatFilterTypeR\x10\x63hat_filter_type\"\xd1\x02\n\x0b\x43hatRequest\x12\x33\n\x08sly_data\x18\x03 \x01(\x0b\x32\x17.google.protobuf.StructR\x08sly_data\x12Y\n\x0cuser_message\x18\x04 \x01(\x0b\x32\x35.dev.cognizant_ai.neuro_san.api.grpc.chat.ChatMessageR\x0cuser_message\x12Y\n\x0c\x63hat_context\x18\x05 \x01(\x0b\x32\x35.dev.cognizant_ai.neuro_san.api.grpc.chat.ChatContextR\x0c\x63hat_context\x12W\n\x0b\x63hat_filter\x18\x06 \x01(\x0b\x32\x35.dev.cognizant_ai.neuro_san.api.grpc.agent.ChatFilterR\x0b\x63hat_filter\"\xaa\x01\n\x0c\x43hatResponse\x12G\n\x07request\x18\x01 \x01(\x0b\x32\x36.dev.cognizant_ai.neuro_san.api.grpc.agent.ChatRequest\x12Q\n\x08response\x18\x04 \x01(\x0b\x32\x35.dev.cognizant_ai.neuro_san.api.grpc.chat.ChatMessageR\x08response\"\x15\n\x13\x43onnectivityRequest\"@\n\x10\x43onnectivityInfo\x12\x16\n\x06origin\x18\x01 \x01(\tR\x06origin\x12\x14\n\x05tools\x18\x02 \x03(\tR\x05tools\"\x81\x01\n\x14\x43onnectivityResponse\x12i\n\x11\x63onnectivity_info\x18\x01 \x03(\x0b\x32;.dev.cognizant_ai.neuro_san.api.grpc.agent.ConnectivityInfoR\x11\x63onnectivity_info*7\n\x0e\x43hatFilterType\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07MINIMAL\x10\x01\x12\x0b\n\x07MAXIMAL\x10\x02\x32\xad\x04\n\x0c\x41gentService\x12\xaa\x01\n\x08\x46unction\x12:.dev.cognizant_ai.neuro_san.api.grpc.agent.FunctionRequest\x1a;.dev.cognizant_ai.neuro_san.api.grpc.agent.FunctionResponse\"%\x82\xd3\xe4\x93\x02\x1f\x12\x1d/api/v1/{agent_name}/function\x12\xb2\x01\n\rStreamingChat\x12\x36.dev.cognizant_ai.neuro_san.api.grpc.agent.ChatRequest\x1a\x37.dev.cognizant_ai.neuro_san.api.grpc.agent.ChatResponse\".\x82\xd3\xe4\x93\x02(\"#/api/v1/{agent_name}/streaming_chat:\x01*0\x01\x12\xba\x01\n\x0c\x43onnectivity\x12>.dev.cognizant_ai.neuro_san.api.grpc.agent.ConnectivityRequest\x1a?.dev.cognizant_ai.neuro_san.api.grpc.agent.ConnectivityResponse\")\x82\xd3\xe4\x93\x02#\x12!/api/v1/{agent_name}/connectivityB^Z\\github.com/leaf-ai/neuro_san/internal/gen/dev.cognizant_ai/neuro_san/api/grpc/agent/v1;agentb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'neuro_san.api.grpc.agent_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z\\github.com/leaf-ai/neuro_san/internal/gen/dev.cognizant_ai/neuro_san/api/grpc/agent/v1;agent'
  _globals['_AGENTSERVICE'].methods_by_name['Function']._options = None
  _globals['_AGENTSERVICE'].methods_by_name['Function']._serialized_options = b'\202\323\344\223\002\037\022\035/api/v1/{agent_name}/function'
  _globals['_AGENTSERVICE'].methods_by_name['StreamingChat']._options = None
  _globals['_AGENTSERVICE'].methods_by_name['StreamingChat']._serialized_options = b'\202\323\344\223\002(\"#/api/v1/{agent_name}/streaming_chat:\001*'
  _globals['_AGENTSERVICE'].methods_by_name['Connectivity']._options = None
  _globals['_AGENTSERVICE'].methods_by_name['Connectivity']._serialized_options = b'\202\323\344\223\002#\022!/api/v1/{agent_name}/connectivity'
  _globals['_CHATFILTERTYPE']._serialized_start=1242
  _globals['_CHATFILTERTYPE']._serialized_end=1297
  _globals['_FUNCTIONREQUEST']._serialized_start=168
  _globals['_FUNCTIONREQUEST']._serialized_end=185
  _globals['_FUNCTION']._serialized_start=187
  _globals['_FUNCTION']._serialized_end=288
  _globals['_FUNCTIONRESPONSE']._serialized_start=290
  _globals['_FUNCTIONRESPONSE']._serialized_end=389
  _globals['_CHATFILTER']._serialized_start=391
  _globals['_CHATFILTER']._serialized_end=506
  _globals['_CHATREQUEST']._serialized_start=509
  _globals['_CHATREQUEST']._serialized_end=846
  _globals['_CHATRESPONSE']._serialized_start=849
  _globals['_CHATRESPONSE']._serialized_end=1019
  _globals['_CONNECTIVITYREQUEST']._serialized_start=1021
  _globals['_CONNECTIVITYREQUEST']._serialized_end=1042
  _globals['_CONNECTIVITYINFO']._serialized_start=1044
  _globals['_CONNECTIVITYINFO']._serialized_end=1108
  _globals['_CONNECTIVITYRESPONSE']._serialized_start=1111
  _globals['_CONNECTIVITYRESPONSE']._serialized_end=1240
  _globals['_AGENTSERVICE']._serialized_start=1300
  _globals['_AGENTSERVICE']._serialized_end=1857
# @@protoc_insertion_point(module_scope)



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

from langchain_community.document_loaders import WebBaseLoader
from langchain_community.vectorstores import InMemoryVectorStore
from langchain_core.documents import Document
from langchain_core.vectorstores.base import VectorStoreRetriever
from langchain_openai import OpenAIEmbeddings
from langchain_text_splitters import RecursiveCharacterTextSplitter

from neuro_san.interfaces.coded_tool import CodedTool


class Rag(CodedTool):
    """
    A CodedTool implementation that performs Retrieval-Augmented Generation (RAG)
    using URLs obtained from a website search.

    This class assumes that a prior search has been conducted to identify relevant URLs.
    It then retrieves the content from those URLs and applies a language model to
    generate responses based on the retrieved information.

    This is useful in workflows where external search tools provide candidate links,
    and RAG is applied afterward to synthesize a meaningful answer from the linked content.
    """

    def invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]):
        """
        :param args: An argument dictionary whose keys are the parameters
                to the coded tool and whose values are the values passed for them
                by the calling agent.  This dictionary is to be treated as read-only.

        :param sly_data: A dictionary whose keys are defined by the agent hierarchy,
                but whose values are meant to be kept out of the chat stream.

                This dictionary is largely to be treated as read-only.
                It is possible to add key/value pairs to this dict that do not
                yet exist as a bulletin board, as long as the responsibility
                for which coded_tool publishes new entries is well understood
                by the agent chain implementation and the coded_tool implementation
                adding the data is not invoke()-ed more than once.

                Keys expected for this implementation are:
                    None
        """

    async def async_invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> str:
        """
        Load documents from URLs, build a vector store, and run a query against it.

        :param args: Dictionary containing 'urls' (list of URLs) and 'query' (search string)
        :param sly_data: A dictionary whose keys are defined by the agent hierarchy,
            but whose values are meant to be kept out of the chat stream.

            This dictionary is largely to be treated as read-only.
            It is possible to add key/value pairs to this dict that do not
            yet exist as a bulletin board, as long as the responsibility
            for which coded_tool publishes new entries is well understood
            by the agent chain implementation and the coded_tool implementation
            adding the data is not invoke()-ed more than once.

            Keys expected for this implementation are:
                None
        :return: Text result from querying the built vector store, or error message
        """
        # Extract arguments from the input dictionary
        urls: List[str] = args.get("urls")
        query: str = args.get("query", "")

        # Validate presence of required inputs
        if not urls:
            return "Error: No URLs provided."
        if not query:
            return "Error: No query provided."

        # Build the vector store and run the query
        vectorstore: InMemoryVectorStore = await self.generate_vectorstore(urls)
        return await self.query_vectorstore(vectorstore, query)

    async def generate_vectorstore(self, urls: List[str]) -> InMemoryVectorStore:
        """
        Asynchronously loads web documents from given URLs, split them into chunks,
        and build an in-memory vector store using OpenAI embeddings.

        :param urls: List of URLs to fetch and embed
        :return: In-memory vector store containing the embedded document chunks
        """

        # Concurrently load documents from all URLs
        loader = WebBaseLoader(urls)
        docs = []
        async for doc in loader.alazy_load():
            docs.append(doc)

        # Split documents into smaller chunks for better embedding and retrieval
        text_splitter = RecursiveCharacterTextSplitter.from_tiktoken_encoder(
            chunk_size=100, chunk_overlap=50
        )
        doc_chunks: List[Document] = text_splitter.split_documents(docs)

        # Create an in-memory vector store with embeddings
        vectorstore: InMemoryVectorStore = await InMemoryVectorStore.afrom_documents(
            documents=doc_chunks,
            collection_name="rag-in-memory",
            embedding=OpenAIEmbeddings(),
        )

        return vectorstore

    async def query_vectorstore(self, vectorstore: InMemoryVectorStore, query: str) -> str:
        """
        Query the given vector store using the provided query string
        and return the combined content of retrieved documents.

        :param vectorstore: The in-memory vector store to query
        :param query: The user query to search for relevant documents
        :return: Concatenated text content of the retrieved documents
        """
        # Create a retriever interface from the vector store
        retriever: VectorStoreRetriever = vectorstore.as_retriever()

        # Perform an asynchronous similarity search
        results: List[Document] = await retriever.ainvoke(query)

        # Concatenate the content of all retrieved documents
        return "\n\n".join(doc.page_content for doc in results)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from collections.abc import Sequence
from typing import Any
from typing import Dict

from pydantic import ConfigDict

from langchain_core.agents import AgentAction
from langchain_core.agents import AgentFinish
from langchain_core.callbacks.base import AsyncCallbackHandler
from langchain_core.documents import Document
from langchain_core.outputs import LLMResult
from langchain_core.outputs.chat_generation import ChatGeneration

from neuro_san.internals.journals.originating_journal import OriginatingJournal
from neuro_san.internals.messages.agent_message import AgentMessage


# pylint: disable=too-many-ancestors
class JournalingCallbackHandler(AsyncCallbackHandler):
    """
    AsyncCallbackHandler implementation that intercepts agent-level chatter

    We use this guy to intercept agent-level messages like:
        "Thought: Do I need a tool?" and preliminary results from the agent

    We are currently only listening to on_llm_end(), but there are many other
    callbacks to hook into, most of which are not really productive.
    Some are overriden here to explore, others are not.  See the base
    AsyncCallbackHandler to explore more.

    Of note: This CallbackHandler mechanism is the kind of thing that
            LoggingCallbackHandler hooks into to produce egregiously chatty logs.
    """

    # Declarations of member variables here satisfy Pydantic style,
    # which is a type validator that langchain is based on which
    # is able to use JSON schema definitions to validate fields.
    journal: OriginatingJournal

    # This guy needs to be a pydantic class and in order to have
    # a non-pydantic Journal as a member, we need to do this.
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(self, journal: OriginatingJournal):
        """
        Constructor

        :param journal: The journal to write messages to
        """
        self.journal: OriginatingJournal = journal

    async def on_llm_end(self, response: LLMResult,
                         **kwargs: Any) -> None:
        # Empirically we have seen that LLMResults that come in on_llm_end() calls
        # have a generations field which is a list of lists. Inside that inner list,
        # the first object is a ChatGeneration, whose text field tends to have agent
        # thinking in it.
        generations = response.generations[0]
        first_generation = generations[0]
        if isinstance(first_generation, ChatGeneration):
            content: str = first_generation.text
            if content is not None and len(content) > 0:
                # Package up the thinking content as an AgentMessage to stream
                message = AgentMessage(content=content.strip())
                # Some AGENT messages that come from this source end up being dupes
                # of AI messages that can come later.
                # Use this method to put the message on hold for later comparison.
                await self.journal.write_message_if_next_not_dupe(message)

    async def on_chain_end(self, outputs: Dict[str, Any],
                           **kwargs: Any) -> None:
        # print(f"In on_chain_end() with {outputs}")
        return

    async def on_tool_end(self, output: Any,
                          **kwargs: Any) -> None:
        # print(f"In on_tool_end() with {output}")
        return

    async def on_agent_action(self, action: AgentAction,
                              **kwargs: Any) -> None:
        # print(f"In on_agent_action() with {action}")
        return

    async def on_agent_finish(self, finish: AgentFinish,
                              **kwargs: Any) -> None:
        # print(f"In on_agent_finish() with {finish}")
        return

    async def on_retriever_end(self, documents: Sequence[Document],
                               **kwargs: Any) -> None:
        # print(f"In on_retriever_end() with {documents}")
        return




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""

import copy
import threading
from typing import Any, Dict, List

from tornado.ioloop import IOLoop

from neuro_san.http_sidecar.logging.http_logger import HttpLogger
from neuro_san.service.agent_server import DEFAULT_FORWARDED_REQUEST_METADATA
from neuro_san.http_sidecar.http_server_app import HttpServerApp

from neuro_san.http_sidecar.interfaces.agent_authorizer import AgentAuthorizer
from neuro_san.http_sidecar.interfaces.agents_updater import AgentsUpdater
from neuro_san.http_sidecar.handlers.health_check_handler import HealthCheckHandler
from neuro_san.http_sidecar.handlers.connectivity_handler import ConnectivityHandler
from neuro_san.http_sidecar.handlers.function_handler import FunctionHandler
from neuro_san.http_sidecar.handlers.streaming_chat_handler import StreamingChatHandler
from neuro_san.http_sidecar.handlers.concierge_handler import ConciergeHandler
from neuro_san.http_sidecar.handlers.openapi_publish_handler import OpenApiPublishHandler


class HttpSidecar(AgentAuthorizer, AgentsUpdater):
    """
    Class provides simple http endpoint for neuro-san API,
    working as a client to neuro-san gRPC service.
    """
    # pylint: disable=too-many-arguments, too-many-positional-arguments
    # pylint: disable=too-many-instance-attributes
    def __init__(self, port: int, http_port: int,
                 agents: Dict[str, Any],
                 openapi_service_spec_path: str,
                 forwarded_request_metadata: str = DEFAULT_FORWARDED_REQUEST_METADATA):
        """
        Constructor:
        :param port: port for gRPC neuro-san service;
        :param http_port: port for http neuro-san service;
        :param agents: dictionary of registered agents;
        :param openapi_service_spec_path: path to a file with OpenAPI service specification;
        :param forwarded_request_metadata: A space-delimited list of http metadata request keys
               to forward to logs/other requests
        """
        self.server_name_for_logs: str = "Http Server"
        self.port = port
        self.http_port = http_port
        self.agents = copy.deepcopy(agents)
        self.logger = None
        self.openapi_service_spec_path: str = openapi_service_spec_path
        self.forwarded_request_metadata: List[str] = forwarded_request_metadata.split(" ")
        self.allowed_agents: Dict[str, bool] = {}
        self.lock = None

    def __call__(self):
        """
        Method to be called by a process running tornado HTTP server
        to actually start serving requests.
        """
        self.lock = threading.Lock()
        self.logger = HttpLogger(self.forwarded_request_metadata)

        app = self.make_app()
        app.listen(self.http_port)
        self.logger.info({}, "HTTP server is running on port %d", self.http_port)
        self.logger.debug({}, "Serving agents: %s", repr(self.agents.keys()))
        # Put agent names into "allowed" list:
        self.allowed_agents = {}
        for agent_name, _ in self.agents.items():
            self.allowed_agents[agent_name] = True
        IOLoop.current().start()

    def make_app(self):
        """
        Construct tornado HTTP "application" to run.
        """
        request_data: Dict[str, Any] = self.build_request_data()
        handlers = []
        handlers.append(("/", HealthCheckHandler))
        handlers.append(("/api/v1/list", ConciergeHandler, request_data))
        handlers.append(("/api/v1/docs", OpenApiPublishHandler, request_data))

        # Register templated request paths for agent API methods:
        # regexp format used here is that of Python Re standard library.
        handlers.append((r"/api/v1/([^/]+)/function", FunctionHandler, request_data))
        handlers.append((r"/api/v1/([^/]+)/connectivity", ConnectivityHandler, request_data))
        handlers.append((r"/api/v1/([^/]+)/streaming_chat", StreamingChatHandler, request_data))

        return HttpServerApp(handlers)

    def allow(self, agent_name) -> bool:
        return self.allowed_agents.get(agent_name, False)

    def update_agents(self, agents: List[str]):
        with self.lock:
            # We assume all agents from "agents" dictionary are enabled:
            for agent_name in agents:
                self.allowed_agents[agent_name] = True
            # All other agents are disabled:
            for agent_name, _ in self.allowed_agents.items():
                if agent_name not in agents:
                    self.allowed_agents[agent_name] = False

    def build_request_data(self) -> Dict[str, Any]:
        """
        Build request data for Http handlers.
        :return: a dictionary with request data to be passed to a http handler.
        """
        return {
            "agent_policy": self,
            "agents_updater": self,
            "port": self.port,
            "forwarded_request_metadata": self.forwarded_request_metadata,
            "openapi_service_spec_path": self.openapi_service_spec_path
        }




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from langchain_core.messages.ai import AIMessage
from langchain_core.messages.base import BaseMessage

from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.agent_tool_result_message import AgentToolResultMessage
from neuro_san.internals.messages.message_utils import is_relevant_to_chat_history


class OriginatingJournal(Journal):
    """
    A Journal implementation that has an origin.
    """

    def __init__(self, wrapped_journal: Journal,
                 origin: List[Dict[str, Any]],
                 chat_history: List[BaseMessage] = None):
        """
        Constructor

        :param wrapped_journal: The Journal that this implementation wraps
        :param origin: The origin that will be applied to all messages.
        :param chat_history: The chat history list instance to store write_message() results in.
                            Can be None (the default).
        """
        self.wrapped_journal: Journal = wrapped_journal
        self.origin: List[Dict[str, Any]] = origin
        self.chat_history: List[BaseMessage] = chat_history
        self.pending: BaseMessage = None

    def get_origin(self) -> List[Dict[str, Any]]:
        """
        :return: The origin associated with this Journal
        """
        return self.origin

    async def write_message(self, message: BaseMessage, origin: List[Dict[str, Any]] = None):
        """
        Writes a BaseMessage entry into the wrapped_journal
        and appends to the chat history.

        :param message: The BaseMessage instance to write to the wrapped_journal
        :param origin: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
                For this particular implementation we expect this to be None
        """
        use_origin: List[Dict[str, Any]] = self.origin
        if origin is not None:
            use_origin = origin

        if self.chat_history is not None and is_relevant_to_chat_history(message):
            # Different LLM providers handle message types differently when constructing responses:
            #
            # - Anthropic models (via ChatAnthropic) explicitly check the `message.type` string
            #   and only accept messages of type "human" or "ai". Custom subclasses like
            #   AgentToolResultMessage return a different type (e.g., "agent_tool_result"),
            #   which causes Anthropic's handler to reject the message.
            #
            #
            # - OpenAI and Ollama models (via ChatOpenAI and ChatOllama) do not rely on `message.type`.
            #   Instead, they use `isinstance(message, AIMessage)` checks, which allows us to safely pass
            #   `AgentToolResultMessage` since it subclasses `AIMessage`. This gives us the flexibility
            #   to include additional metadata like `tool_result_origin` when supported.
            #
            # To avoid problem with any other LLMs, convert "AgentToolResultMessage" to "AIMessage"
            # when appending it chathistory but allow it to be written in the journal as is to
            # to maintain the information on tool origin.
            if isinstance(message, AgentToolResultMessage):
                ai_message = AIMessage(content=message.content)
                self.chat_history.append(ai_message)
            else:
                self.chat_history.append(message)

        if self.pending is not None:
            # Avoid cases where two different kinds of message hold the same content.
            if self.pending.content != message.content:
                await self.wrapped_journal.write_message(self.pending, use_origin)
            self.pending = None

        await self.wrapped_journal.write_message(message, use_origin)

    def get_chat_history(self) -> List[BaseMessage]:
        """
        :return: The chat history list of base messages associated with the instance.
        """
        return self.chat_history

    async def write_message_if_next_not_dupe(self, message: BaseMessage):
        """
        Writes a BaseMessage entry into the wrapped_journal
        and appends to the chat history as long as the next message does not have the same content.

        :param message: The BaseMessage instance to write to the wrapped_journal
        """
        if self.pending is not None:
            # Flush if anything is already waiting.
            await self.write_message(self.pending)
        self.pending = message




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import List


class AgentsUpdater:
    """
    Abstract interface for updating current collection of agents
    being served.
    """

    def update_agents(self, agents: List[str]):
        """
        :param agents: list of agents names which should be served currently.
        :return: nothing
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import AsyncGenerator

from leaf_common.session.async_abstract_service_session import AsyncAbstractServiceSession
from leaf_common.time.timeout import Timeout

from neuro_san.api.grpc import agent_pb2 as service_messages
from neuro_san.interfaces.async_agent_session import AsyncAgentSession
from neuro_san.session.agent_service_stub import AgentServiceStub


class AsyncGrpcServiceAgentSession(AsyncAbstractServiceSession, AsyncAgentSession):
    """
    Implementation of AsyncAgentSession that talks to a gRPC service asynchronously.
    """

    DEFAULT_PORT: int = AsyncAgentSession.DEFAULT_PORT
    DEFAULT_AGENT_NAME: str = "esp_decision_assistant"

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self, host: str = None,
                 port: str = None,
                 timeout_in_seconds: int = 30,
                 metadata: Dict[str, str] = None,
                 security_cfg: Dict[str, Any] = None,
                 umbrella_timeout: Timeout = None,
                 streaming_timeout_in_seconds: int = None,
                 agent_name: str = DEFAULT_AGENT_NAME):
        """
        Creates an AsyncAgentSession that connects to the
        Agent Service and delegates its implementations to the service.

        :param host: the service host to connect to
                        If None, will use a default
        :param port: the service port
                        If None, will use a default
        :param timeout_in_seconds: timeout to use when communicating
                        with the service
        :param metadata: A grpc metadata of key/value pairs to be inserted into
                         the header. Default is None. Preferred format is a
                         dictionary of string keys to string values.
        :param security_cfg: An optional dictionary of parameters used to
                        secure the TLS and the authentication of the gRPC
                        connection.  Supplying this implies use of a secure
                        GRPC Channel.  Default is None, uses insecure channel.
        :param umbrella_timeout: A Timeout object under which the length of all
                        looping and retries should be considered
        :param streaming_timeout_in_seconds: timeout to use when streaming to/from
                        the service. Default is None, indicating connection should
                        stay open until the (last) result is yielded.
        :param agent_name: The name of the agent to talk to
        """
        use_host: str = "localhost"
        if host is not None:
            use_host = host

        use_port: str = str(self.DEFAULT_PORT)
        if port is not None:
            use_port = port

        # Normally we pass around the service_stub like a class,
        # but AgentServiceStub has a __call__() method to intercept
        # constructor-like behavior.
        service_stub = AgentServiceStub(agent_name)
        AsyncAbstractServiceSession.__init__(self, "Agent Server",
                                             service_stub,
                                             use_host, use_port,
                                             timeout_in_seconds, metadata,
                                             security_cfg, umbrella_timeout,
                                             streaming_timeout_in_seconds)

    async def function(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the FunctionRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the FunctionResponse
                    protobufs structure. Has the following keys:
                "function" - the dictionary description of the function
        """
        # pylint: disable=no-member
        return await self.call_grpc_method(
            "function",
            self._function_from_stub,
            request_dict,
            service_messages.FunctionRequest())

    async def connectivity(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the ConnectivityRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the ConnectivityResponse
                    protobufs structure. Has the following keys:
                "connectivity_info" - the list of connectivity descriptions for
                                    each node in the agent network the service
                                    wants the client ot know about.
        """
        # pylint: disable=no-member
        return await self.call_grpc_method(
            "connectivity",
            self._connectivity_from_stub,
            request_dict,
            service_messages.ConnectivityRequest())

    async def streaming_chat(self, request_dict: Dict[str, Any]) -> AsyncGenerator[Dict[str, Any], None]:
        """
        :param request_dict: A dictionary version of the ChatRequest
                    protobufs structure. Has the following keys:
            "user_message"  - A ChatMessage dict representing the user input to the chat stream
            "chat_context"  - A ChatContext dict

        :return: An iterator of dictionary versions of the ChatResponse
                    protobufs structure. Has the following keys:
            "response"      - An optional ChatMessage dictionary.  See chat.proto for details.

            Note that responses to the chat input might be numerous and will come as they
            are produced until the system decides there are no more messages to be sent.
        """
        # pylint: disable=no-member
        generator = self.stream_grpc_method(
            "streaming_chat",
            self._streaming_chat_from_stub,
            request_dict,
            request_instance=service_messages.ChatRequest())

        # Cannot do "yield from" in async land. Have to make explicit loop
        async for response in generator:
            yield response

    @staticmethod
    async def _function_from_stub(stub, timeout_in_seconds,
                                  metadata, credentials, *args):
        """
        Global method associated with the session that calls Function
        given a grpc Stub already set up with a channel (socket) to call with.
        """
        response = await stub.Function(*args, timeout=timeout_in_seconds,
                                       metadata=metadata,
                                       credentials=credentials)
        return response

    @staticmethod
    async def _connectivity_from_stub(stub, timeout_in_seconds,
                                      metadata, credentials, *args):
        """
        Global method associated with the session that calls Connectivity
        given a grpc Stub already set up with a channel (socket) to call with.
        """
        response = await stub.Connectivity(*args, timeout=timeout_in_seconds,
                                           metadata=metadata,
                                           credentials=credentials)
        return response

    @staticmethod
    async def _streaming_chat_from_stub(stub, timeout_in_seconds,
                                        metadata, credentials, *args):
        """
        Global method associated with the session that calls StreamingChat
        given a grpc Stub already set up with a channel (socket) to call with.
        """
        generator = stub.StreamingChat(*args, timeout=timeout_in_seconds,
                                       metadata=metadata,
                                       credentials=credentials)
        # Cannot do "yield from" in async land. Have to make explicit loop
        async for response in generator:
            yield response




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.run_context.factory.master_llm_factory import MasterLlmFactory
from neuro_san.internals.run_context.interfaces.run_context import RunContext
from neuro_san.internals.run_context.interfaces.tool_caller import ToolCaller
from neuro_san.internals.run_context.langchain.langchain_run_context import LangChainRunContext
from neuro_san.internals.run_context.openai.openai_run_context import OpenAIRunContext


class RunContextFactory:
    """
    Creates the correct kind of RunContext
    """

    @staticmethod
    def create_run_context(parent_run_context: RunContext,
                           tool_caller: ToolCaller,
                           invocation_context: InvocationContext = None,
                           chat_context: Dict[str, Any] = None,
                           config: Dict[str, Any] = None) \
            -> RunContext:
        """
        Creates an appropriate RunContext

        :param parent_run_context: The parent RunContext (if any) to pass
                             down its resources to a new RunContext created by
                             this call.
        :param tool_caller: The ToolCaller whose lifespan matches that
                            of the newly created RunContext
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        :param config: The config dictionary which may or may not contain
                       keys for the context_type and default llm_config
        """

        # Initialize return value
        run_context: RunContext = None

        empty: Dict[str, Any] = {}
        use_config: Dict[str, Any] = config
        if use_config is None:
            use_config = empty

        # Get some fields from the config with reasonable defaults
        default_llm_config: Dict[str, Any] = {
            "model_name": "gpt-4o",
            "verbose": False
        }
        default_llm_config = use_config.get("llm_config", default_llm_config)

        # Prepare for sanity in checks below
        context_type: str = MasterLlmFactory.get_context_type(use_config)

        use_invocation_context: InvocationContext = invocation_context
        if use_invocation_context is None and parent_run_context is not None:
            use_invocation_context = parent_run_context.get_invocation_context()

        if context_type.startswith("openai"):
            run_context = OpenAIRunContext(default_llm_config, parent_run_context,
                                           tool_caller, use_invocation_context,
                                           chat_context)
        elif context_type.startswith("langchain"):
            run_context = LangChainRunContext(default_llm_config, parent_run_context,
                                              tool_caller, use_invocation_context,
                                              chat_context)
        else:
            # Default case
            run_context = LangChainRunContext(default_llm_config, parent_run_context,
                                              tool_caller, use_invocation_context,
                                              chat_context)

        return run_context




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

import json

from openai.types.beta.threads.required_action_function_tool_call \
    import Function
from openai.types.beta.threads.required_action_function_tool_call \
    import RequiredActionFunctionToolCall

from neuro_san.internals.run_context.interfaces.tool_call import ToolCall


class OpenAIToolCall(ToolCall):
    """
    An OpenAI implementation for ToolCall.
    """

    def __init__(self, tool_call: RequiredActionFunctionToolCall):
        """
        Constructor
        :param tool_call: The OpenAI RequiredActionFunctionToolCall on which this
                          implementation is based.
        """
        self.tool_call = tool_call

    def get_id(self) -> str:
        """
        :return: The string id of this run
        """
        return self.tool_call.id

    def get_function_arguments(self) -> Dict[str, Any]:
        """
        :return: Returns a dictionary of the function arguments for the tool call
        """
        function: Function = self.tool_call.function
        arguments: str = function.arguments

        function_arguments: Dict[str, Any] = json.loads(arguments)

        return function_arguments

    def get_function_name(self) -> str:
        """
        :return: Returns the string name of the tool
        """
        function: Function = self.tool_call.function
        return function.name




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from neuro_san.internals.filters.answer_message_filter import AnswerMessageFilter
from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.message_processing.message_processor import MessageProcessor


class AnswerMessageProcessor(MessageProcessor):
    """
    Implementation of the MessageProcessor that looks for "the answer"
    of the chat session.
    """

    def __init__(self):
        """
        Constructor
        """
        self.answer: str = None
        self.answer_origin: List[Dict[str, Any]] = None
        self.filter: AnswerMessageFilter = AnswerMessageFilter()

    def get_answer(self) -> str:
        """
        :return: The final answer from the agent session interaction
        """
        return self.answer

    def get_answer_origin(self) -> List[Dict[str, Any]]:
        """
        :return: The origin of the final answer from the agent session interaction
        """
        return self.answer_origin

    def reset(self):
        """
        Resets any previously accumulated state
        """
        self.answer = None
        self.answer_origin = None

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """
        if not self.filter.allow_message(chat_message_dict, message_type):
            # Does not pass the criteria for a message holding a final answer
            return

        origin: List[Dict[str, Any]] = chat_message_dict.get("origin")
        text = chat_message_dict.get("text")

        # Record what we got.
        # We might get another as we go along, but the last message in the stream
        # meeting the criteria above is our final answer.
        if text is not None:
            self.answer = text
        if origin is not None:
            self.answer_origin = origin




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from inspect import signature
from types import MethodType
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Type
from typing import Union

import os

from langchain.tools import BaseTool
from langchain_community.agent_toolkits.base import BaseToolkit
from pydantic import BaseModel

from leaf_common.config.dictionary_overlay import DictionaryOverlay
from leaf_common.config.resolver import Resolver

from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.run_context.langchain.toolbox_info_restorer import ToolboxInfoRestorer


class ToolboxFactory(ContextTypeToolboxFactory):
    """
    A factory class for creating instances of various tools defined in the toolbox.

    This class provides an interface to instantiate different tools based on the specified langchain base tools
    and predefined coded tools.

    This approach standardizes tool creation and simplifies integration with agents requiring predefined tools.

    ### Supported Tools:

    #### Langchain Tools:

    - **Search Tools:**
        - "bing_search": Returns a "BingSearchResults" instance for performing Bing search queries.
        - "tavily_search": Returns a "TavilySearchResults" instance for performing Tavily search queries.

    - **HTTP Request Tools:** These allow making different types of HTTP requests using the "RequestsToolkit"
      with a "TextRequestsWrapper". The following tool names are available:
        - "requests_get": For making GET requests.
        - "requests_post": For making POST requests.
        - "requests_patch": For making PATCH requests.
        - "requests_put": For making PUT requests.
        - "requests_delete": For making DELETE requests.
        - "requests_toolkit": For all of the above request tools.

    #### Coded Tools:
        - "website_search": Internet search based on DuckDuckGo Search
        - "rag_retriever": Perfrom retrieval-augmented generation on given urls

    ### Extending the Class

        To integrate additional tools, add a tool configuration file in JSON or HOCON format
        and set its path to the environment variable "AGENT_TOOLBOX_INFO_FILE".

        The configuration should follow this structure

        for langchain's tools:
        - The tool name serves as a key.
        - The corresponding value should be a dictionary with:
        - "class": The fully qualified class name of the tool.
        - "args": A dictionary of arguments required for the tool's initialization,
            which may include nested class configurations.

        for coded tools:
        - The tool name serves as a key.
        - The corresponding value should be a dictionary with:
        - "class": Module and class in the format of tool_module.ClassName where tool_module is in
                    AGENT_TOOL_PATH or neuro_san/coded_tools.
        - "description": When and how to use the tool.
        - "parameters": Information on arguments of the tool.
            See "parameters" in https://github.com/leaf-ai/neuro-san/blob/main/docs/agent_hocon_reference.md

        The default toolbox config file can be seen at
        "neuro_san/internals/run_context/langchain/toolbox_info.hocon"
    """

    def __init__(self):
        """
        Constructor
        """
        self.toolbox_infos: Dict[str, Any] = {}
        self.overlayer = DictionaryOverlay()
        self.toolbox_info_file: str = None

    def load(self):
        """
        Loads the base tool information from hocon files.
        """
        restorer = ToolboxInfoRestorer()
        self.toolbox_infos = restorer.restore()

        # Mix in user-specified toolbox info, if available.
        toolbox_info_file: str = os.getenv("AGENT_TOOLBOX_INFO_FILE")
        if toolbox_info_file is not None and len(toolbox_info_file) > 0:
            extra_toolbox_infos: Dict[str, Any] = restorer.restore(file_reference=toolbox_info_file)
            self.toolbox_infos = self.overlayer.overlay(self.toolbox_infos, extra_toolbox_infos)

            self.toolbox_info_file = toolbox_info_file

    def create_tool_from_toolbox(
            self,
            tool_name: str,
            user_args: Dict[str, Any] = None
    ) -> Union[BaseTool, Dict[str, Any], List[BaseTool]]:
        """
        Resolves dependencies and instantiates the requested tool.

        :param tool_name: The name of the tool to instantiate.
        :param user_args: Arguments provided by the user, which override the config file.
        :return: - Instantiated tool if "class" of tool_name points to a BaseTool class
                 - A list of tools if "class of "tool_name points to a BaseToolkit class.
                 - A dict of tool's "description" and "parameters" if tool_name points to a CodedTool
        """
        empty: Dict[str, Any] = {}

        tool_info: Dict[str, Any] = self.toolbox_infos.get(tool_name)
        if not tool_info:
            raise ValueError(f"Tool '{tool_name}' is not defined in {self.toolbox_info_file}.")

        if not isinstance(tool_info, Dict):
            raise ValueError(f"The value for the {tool_name} key must be a dictionary.")

        if "class" not in tool_info:
            raise ValueError(
                "Missing required key: 'class'.\n"
                "Each tool must include a 'class' key:\n"
                "- For Langchain base tools: use the full class path "
                "(e.g., 'langchain_community.tools.bing_search.BingSearchResults')\n"
                "- For shared CodedTools: use 'module.Class' format (e.g., 'websearch.WebSearch')"
            )

        # If "description" in the tool info, then it is a shared coded tool.
        # Return dictionary of tool's description and parameters.
        if "description" in tool_info:
            return tool_info

        # Instantiate the main tool or toolkit class
        tool_class: Type[Any] = self._resolve_class(tool_info.get("class"))
        # Recursively resolve arguments (including wrapper dependencies)
        resolved_args: Dict[str, Any] = self._resolve_args(tool_info.get("args", empty))
        # Merge with user arguments where user_args get the priority
        final_args: Dict[str, Any] = self.overlayer.overlay(resolved_args, user_args) if user_args else resolved_args

        # Use the "from_{tool_name}_api_wrapper" method if available, otherwise the constructor
        callable_obj: Union[Type[BaseTool], Type[BaseToolkit], Callable[..., Any]] = \
            self._get_from_api_wrapper_method(tool_class) or tool_class

        # Validate and instantiate
        self._check_invalid_args(callable_obj, final_args)
        # Instance can be a BaseTool or a BaseToolkit
        instance: Union[BaseTool, BaseToolkit] = callable_obj(**final_args)

        # If the instantiated class has "get_tools()", assume it's a toolkit and return a list of tools
        if hasattr(instance, "get_tools") and callable(instance.get_tools):
            return instance.get_tools()

        return instance

    def _resolve_args(self, args: Dict[str, Any]) -> Dict[str, Any]:
        """
        Recursive resolves arguments when there is a wrapper class as an argument,
        otherwise return args as a dictionary.

        :param args: The arguments to resolve.
        :return: A dictionary of resolved arguments.
        """
        empty: Dict[str, Any] = {}

        resolved_args: Dict[str, Any] = {}
        for key, value in args.items():
            if isinstance(value, dict) and "class" in value:
                # If the argument is a class definition, resolve and instantiate it
                nested_class: BaseModel = self._resolve_class(value.get("class"))
                nested_args: Dict[str, Any] = self._resolve_args(value.get("args", empty))
                self._check_invalid_args(nested_class, nested_args)
                resolved_args[key] = nested_class(**nested_args)
            else:
                # Otherwise, keep primitive values as they are
                resolved_args[key] = value
        return resolved_args

    def _resolve_class(self, class_path: str) -> Type[BaseTool]:
        """
        Uses Resolver to dynamically import a class.

        :param class_path: Full class path (e.g., "package.module.ClassName").
        :return: The resolved class type.
        """
        class_split: List[str] = class_path.split(".")
        if len(class_split) <= 2:
            raise ValueError(
                f"Value in 'class' in {self.toolbox_info_file} must be of the form "
                "'<package_name>.<module_name>.<ClassName>'"
            )

        # Extract module and class details
        packages: List[str] = [".".join(class_split[:-2])]
        class_name: str = class_split[-1]
        resolver = Resolver(packages)

        # Resolve class
        try:
            return resolver.resolve_class_in_module(class_name, module_name=class_split[-2])
        except AttributeError as exception:
            raise ValueError(f"Class {class_path} not found in PYTHONPATH") from exception

    def _check_invalid_args(self, method_class: Union[Type, MethodType], args: Dict[str, Any]):
        """
        Check for invalid arguments in class or method
        :param method_class: Class or method to check for the invalid arguments
        :param args: Arguments to check
        """
        class_args_set: Set[str] = set(signature(method_class).parameters.keys())
        args_set: Set[str] = set(args.keys())
        invalid_args: Set[str] = args_set - class_args_set

        if invalid_args:
            raise ValueError(
                f"Arguments {invalid_args} for '{method_class.__name__}' do not match any attributes "
                "of the class or any arguments of the method."
            )

    def _get_from_api_wrapper_method(
        self,
        tool_class: Union[Type[BaseTool], Type[BaseToolkit]]
    ) -> Optional[Callable[..., Any]]:
        """
        Get a 'from_{tool_name}_api_wrapper' class method from the tool class if available.

        :param tool_class: BaseTool or BaseToolkit class to check for the method.
        :return: The method if found, None otherwise.
        """
        for attr_name in dir(tool_class):
            if attr_name.startswith("from") and attr_name.endswith("api_wrapper"):
                attr: Callable[..., Any] = getattr(tool_class, attr_name)
                if callable(attr):
                    return attr
        return None

    def get_shared_coded_tool_class(self, tool_name: str) -> str:
        """
        Get class of the shared coded tool

        :param tool_name: The name of the tool
        :return: The class of the coded tool
        """
        tool_info: Dict[str, Any] = self.toolbox_infos.get(tool_name)
        return tool_info.get("class")




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any


class AsyncHopper:
    """
    An interface whose clients store things for later use.
    """

    async def put(self, item: Any):
        """
        :param item: The item to put in the hopper.
        """
        raise NotImplementedError



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Dict
from typing import Tuple
from typing import Type

from neuro_san.test.evaluators.gist_agent_evaluator import GistAgentEvaluator
from neuro_san.test.evaluators.greater_agent_evaluator import GreaterAgentEvaluator
from neuro_san.test.evaluators.keywords_agent_evaluator import KeywordsAgentEvaluator
from neuro_san.test.evaluators.less_agent_evaluator import LessAgentEvaluator
from neuro_san.test.evaluators.value_agent_evaluator import ValueAgentEvaluator
from neuro_san.test.interfaces.agent_evaluator import AgentEvaluator
from neuro_san.test.interfaces.assert_forwarder import AssertForwarder


class AgentEvaluatorFactory:
    """
    Factory that creates AgentEvaluators
    """

    NAME_TO_AGENT_EVALUATOR: Dict[str, Tuple[Type[AgentEvaluator], bool]] = {
        "gist": (GistAgentEvaluator, False),
        "not_gist": (GistAgentEvaluator, True),
        "greater": (GreaterAgentEvaluator, False),
        "not_greater": (GreaterAgentEvaluator, True),
        "keywords": (KeywordsAgentEvaluator, False),
        "not_keywords": (KeywordsAgentEvaluator, True),
        "less": (LessAgentEvaluator, False),
        "not_less": (LessAgentEvaluator, True),
        "value": (ValueAgentEvaluator, False),
        "not_value": (ValueAgentEvaluator, True),
    }

    @staticmethod
    def create_evaluator(asserts: AssertForwarder, evaluation_type: str) -> AgentEvaluator:
        """
        Creates AgentEvaluators

        :param asserts: The AssertForwarder instance to handle failures
        :param evaluation_type: A string key describing how the evaluation will take place
        """
        evaluator: AgentEvaluator = None

        # Return early
        if evaluation_type is None:
            return evaluator

        # Look up in the table
        lower_eval: str = evaluation_type.lower()
        eval_tuple: Tuple[Type[AgentEvaluator], bool] = AgentEvaluatorFactory.NAME_TO_AGENT_EVALUATOR.get(lower_eval)
        if eval_tuple is None:
            return evaluator

        # Get components of table value
        eval_class: Type[AgentEvaluator] = eval_tuple[0]
        negate: bool = eval_tuple[1]

        if negate is not None:
            evaluator = eval_class(asserts, negate=negate)
        else:
            evaluator = eval_class(asserts)

        return evaluator




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

import os

from pathlib import Path

from leaf_common.config.dictionary_overlay import DictionaryOverlay
from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.internals.graph.tools.branch_tool import BranchTool
from neuro_san.internals.graph.tools.class_tool import ClassTool
from neuro_san.internals.graph.tools.external_tool import ExternalTool
from neuro_san.internals.graph.tools.toobox_tool import ToolboxTool
from neuro_san.internals.graph.tools.front_man import FrontMan
from neuro_san.internals.graph.tools.sly_data_redactor import SlyDataRedactor
from neuro_san.internals.run_context.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.run_context.interfaces.callable_tool import CallableTool
from neuro_san.internals.run_context.interfaces.run_context import RunContext
from neuro_san.internals.run_context.utils.external_agent_parsing import ExternalAgentParsing
from neuro_san.internals.utils.file_of_class import FileOfClass


class AgentToolRegistry(AgentToolFactory):
    """
    A place where agent tools are registered.
    """

    def __init__(self, config: Dict[str, Any], name: str, agent_tool_path: str = None):
        """
        Constructor

        :param agent_specs: A list of agents to pre-register
        :param name: The name of the registry
        :param agent_tool_path: Optional path to specify where source gets resolved.
                    If None, the value comes from the env var AGENT_TOOL_PATH.
                    If that is not set, if defaults to a path relative to this file.
        """
        self.config = config
        self.name = name
        self.agent_spec_map: Dict[str, Dict[str, Any]] = {}

        self.agent_tool_path = self.determine_agent_tool_path(agent_tool_path)
        self.first_agent: str = None

        agent_specs = self.config.get("tools")
        if agent_specs is not None:
            for agent_spec in agent_specs:
                self.register(agent_spec)

    def determine_agent_tool_path(self, agent_tool_path: str) -> str:
        """
        Policy for determining where tool source should be looked for
        when resolving references to coded tools.

        :return: the agent tool path to use for source resolution.
        """
        # Try the env var first if nothing to start with
        if agent_tool_path is None:
            agent_tool_path = os.environ.get("AGENT_TOOL_PATH")

        # Try reach-around directory if still nothing to start with
        if agent_tool_path is None:
            file_of_class = FileOfClass(__file__, "../../../coded_tools")
            agent_tool_path = file_of_class.get_basis()

        # If we are dealing with file paths, convert that to something resolvable
        if agent_tool_path.find(os.sep) >= 0:

            # Find the best of many resolution paths in the PYTHONPATH
            resolved_tool_path: str = str(Path(agent_tool_path).resolve())
            best_path = ""
            pythonpath: str = os.environ.get("PYTHONPATH")
            if pythonpath is None:
                # Trust what we have already
                best_path = agent_tool_path
            else:
                pythonpath_split = pythonpath.split(":")
                for one_path in pythonpath_split:
                    resolved_path: str = str(Path(one_path).resolve())
                    if resolved_tool_path.startswith(resolved_path) and \
                            len(resolved_path) > len(best_path):
                        best_path = resolved_path

            if len(best_path) == 0:
                raise ValueError(f"No reasonable agent tool path found in PYTHONPATH for {agent_tool_path}")

            # Find the path beneath the python path
            path_split = resolved_tool_path.split(best_path)
            if len(path_split) < 2:
                raise ValueError("""
Cannot find tool path for {agent_tool_path} in PYTHONPATH.
Check to be sure your value for PYTHONPATH includes where you expect where your coded tools live.
""")
            resolve_path = path_split[1]

            # Replace separators with python delimiters for later resolution
            agent_tool_path = resolve_path.replace(os.sep, ".")

            # Remove any leading .s
            while agent_tool_path.startswith("."):
                agent_tool_path = agent_tool_path[1:]

        # Be sure the name of the agent (stem of the hocon file) is the
        # last piece to narrow down the path resolution further.
        if not agent_tool_path.endswith(self.name):
            agent_tool_path = f"{agent_tool_path}.{self.name}"

        return agent_tool_path

    def get_config(self) -> Dict[str, Any]:
        """
        :return: The config dictionary passed into the constructor
        """
        return self.config

    def get_agent_tool_path(self) -> str:
        """
        :return: The path under which tools for this registry should be looked for.
        """
        return self.agent_tool_path

    def register(self, agent_spec: Dict[str, Any]):
        """
        :param agent_spec: A single agent to register
        """
        if agent_spec is None:
            return

        name: str = self.get_name_from_spec(agent_spec)
        if self.first_agent is None:
            self.first_agent = name

        if name in self.agent_spec_map:
            message: str = f"""
The agent named "{name}" appears to have a duplicate entry in its hocon file.
Agent names must be unique within the scope of a single hocon file.

Some things to try:
1. Rename one of the agents named "{name}". Don't forget to scrutinize all the
   tools references from other agents connecting to it.
2. If one definition is an alternate implementation, consider commenting out
   one of them with "#"-style comments.  (Yes, you can do that in a hocon file).
"""
            raise ValueError(message)

        self.agent_spec_map[name] = agent_spec

    def get_name_from_spec(self, agent_spec: Dict[str, Any]) -> str:
        """
        :param agent_spec: A single agent to register
        :return: The agent name as per the spec
        """
        extractor = DictionaryExtractor(agent_spec)
        name = extractor.get("function.name")
        if name is None:
            name = agent_spec.get("name")

        return name

    def get_agent_tool_spec(self, name: str) -> Dict[str, Any]:
        """
        :param name: The name of the agent tool to get out of the registry
        :return: The dictionary representing the spec registered agent
        """
        if name is None:
            return None

        return self.agent_spec_map.get(name)

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def create_agent_tool(self, parent_run_context: RunContext,
                          parent_agent_spec: Dict[str, Any],
                          name: str,
                          sly_data: Dict[str, Any],
                          arguments: Dict[str, Any] = None) -> CallableTool:
        """
        Create an active node for an agent from its spec.

        :param parent_run_context: The RunContext of the agent calling this method
        :param parent_agent_spec: The spec of the agent calling this method.
        :param name: The name of the agent to get out of the registry
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be an empty dictionary.
        :param arguments: A dictionary of arguments for the newly constructed agent
        :return: The CallableTool agent referred to by the name.
        """
        agent_tool: CallableTool = None
        factory: AgentToolFactory = self

        agent_tool_spec: Dict[str, Any] = self.get_agent_tool_spec(name)
        if agent_tool_spec is None:

            if not ExternalAgentParsing.is_external_agent(name):
                raise ValueError(f"No agent_tool_spec for {name}")

            # For external tools, we want to redact the sly data based on
            # the calling/parent's agent specs.
            redacted_sly_data: Dict[str, Any] = self.redact_sly_data(parent_run_context, sly_data)

            # Get the spec for allowing upstream data
            extractor = DictionaryExtractor(parent_agent_spec)
            empty = {}
            allow_from_downstream: Dict[str, Any] = extractor.get("allow.from_downstream", empty)

            agent_tool = ExternalTool(parent_run_context, factory, name, arguments, redacted_sly_data,
                                      allow_from_downstream)
            return agent_tool

        # Merge the arguments coming in from the LLM with those that were specified
        # in the hocon file for the agent.
        use_args: Dict[str, Any] = self.merge_args(arguments, agent_tool_spec)

        if agent_tool_spec.get("toolbox") is not None:
            # If a toolbox is in the spec, this is a shared coded tool where tool's description and
            # args schema are defined in either AGENT_TOOLBOX_INFO_FILE or toolbox_info.hocon.
            agent_tool = ToolboxTool(parent_run_context, factory, use_args, agent_tool_spec, sly_data)
            return agent_tool

        if agent_tool_spec.get("function") is not None:
            # If we have a function in the spec, the agent has arguments
            # it wants to be called with.
            if agent_tool_spec.get("class") is not None:
                # Agent specifically requested a python class to be run,
                # and tool's description and args schema are defined in agent network hocon.
                agent_tool = ClassTool(parent_run_context, factory, use_args, agent_tool_spec, sly_data)
            else:
                agent_tool = BranchTool(parent_run_context, factory, use_args, agent_tool_spec, sly_data)
        else:
            # Get the tool to call from the spec.
            agent_tool = FrontMan(parent_run_context, factory, agent_tool_spec, sly_data)

        return agent_tool

    def create_front_man(self,
                         sly_data: Dict[str, Any] = None,
                         parent_run_context: RunContext = None) -> FrontMan:
        """
        Find and create the FrontMan for chat
        """
        front_man_name: str = self.find_front_man()

        agent_tool_spec: Dict[str, Any] = self.get_agent_tool_spec(front_man_name)
        front_man = FrontMan(parent_run_context, self, agent_tool_spec, sly_data)
        return front_man

    def find_front_man(self) -> str:
        """
        :return: A single tool name to use as the root of the chat agent.
                 This guy will be user facing.  If there are none or > 1,
                 an exception will be raised.
        """
        front_men: List[str] = []

        # Identify the "front-man" agent.
        # Primary heuristic: an agent with defined instructions and a function that takes no parameters.
        # The presence of instructions ensures it was explicitly defined, since users may add parameters
        # to front-men, making function signature alone unreliable.
        for name, agent_spec in self.agent_spec_map.items():
            instructions: str = agent_spec.get("instructions")
            function: Dict[str, Any] = agent_spec.get("function")
            if instructions is not None and function is not None and function.get("parameters") is None:
                front_men.append(name)

        if len(front_men) == 0:
            # The next way to find a front man is to see which agent was registered first
            front_men.append(self.first_agent)

        if len(front_men) == 0:
            raise ValueError("No front man for chat found. "
                             "One entry's function must not have any parameters defined to be the front man")

        if len(front_men) > 1:
            raise ValueError(f"Found > 1 front man for chat. Possibilities: {front_men}")

        front_man = front_men[0]
        return front_man

    def merge_args(self, llm_args: Dict[str, Any], agent_tool_spec: Dict[str, Any]) -> Dict[str, Any]:
        """
        Merges the args specified by the llm with "hard-coded" args specified in the agent spec.
        Hard-coded args win over llm-specified args if both are defined.
        If you want the llm args to win out over the hard-coded args, use a default for
        the function spec instead of the hard-coded args.

        :param llm_args: argument dictionary that the LLM wants
        :param agent_tool_spec: The dictionary representing the spec registered agent
        """
        config_args: Dict[str, Any] = agent_tool_spec.get("args")
        if config_args is None:
            # Nothing to override
            return llm_args

        overlay = DictionaryOverlay()
        merged_args: Dict[str, Any] = overlay.overlay(llm_args, config_args)
        return merged_args

    def redact_sly_data(self, parent_run_context: RunContext, sly_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Redact the sly_data based on the agent spec associated with the parent run context

        :param parent_run_context: The parent run context of the tool to be created.
        :param sly_data: The internal representation of the sly_data to be redacted
        :return: A new sly_data dictionary, redacted as per the parent spec
        """
        parent_spec: Dict[str, Any] = None
        if parent_run_context is not None:
            parent_spec = parent_run_context.get_agent_tool_spec()

        redactor = SlyDataRedactor(parent_spec, config_keys=["allow.sly_data", "allow.to_downstream.sly_data"])
        redacted: Dict[str, Any] = redactor.filter_config(sly_data)
        return redacted

    def get_agent_llm_info_file(self) -> str:
        """
        :return: The absolute path of agent llm info file for llm extension.
        """
        return self.config.get("agent_llm_info_file")




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from copy import deepcopy

from leaf_common.config.config_filter import ConfigFilter


class AbstractCommonDefsConfigFilter(ConfigFilter):
    """
    An abstract ConfigFilter implementation that takes a agent tool registry config
    that may or may not contain a particular set commondefs definitions
    and does an calls an abstract make_replacements() method for appropriate
    values.
    """

    def __init__(self, commondefs_key: str, replacements: Dict[str, Any] = None):
        """
        Constructor

        :param commondefs_key: The string key in the commondefs section
                that contains the replacement dictionary.
        :param replacements: An initial replacements dictionary to start out with
                whose (copied) contents will be updated with commondefs definitions
                in the basis_config during the filter_config() entry point.
                Default is None, indicating everything needed comes from the config.
        """
        self.commondefs_key: str = commondefs_key
        self.starting_replacements: Dict[str, Any] = replacements
        if replacements is None:
            self.starting_replacements = {}

    def filter_config(self, basis_config: Dict[str, Any]) -> Dict[str, Any]:
        """
        Filters the given basis config.

        Ideally this would be a Pure Function in that it would not
        modify the caller's arguments so that the caller has a chance
        to decide whether to take any changes returned.

        :param basis_config: The config dictionary to act as the basis
                for filtering
        :return: A config dictionary, potentially modified as per the
                policy encapsulated by the implementation
        """
        if basis_config is None or not basis_config:    # Empty dictionaries evaluate to False
            # Nothing to modify
            return basis_config

        # Start our replacement dictionary with what was passed into the constructor
        replacements: Dict[str, Any] = deepcopy(self.starting_replacements)

        # Look for something to replace
        commondefs: Dict[str, Any] = {}
        commondefs = basis_config.get("commondefs", commondefs)
        if commondefs:          # Empty dictionaries evaluate to False
            new_replacements: Dict[str, Any] = {}
            new_replacements = deepcopy(commondefs.get(self.commondefs_key, new_replacements))
            replacements.update(new_replacements)

        if not replacements:               # Empty dictionaries evaluate to False
            # No modifications to make
            return basis_config

        # Start out with a copy of the basis.  Leave the input alone.
        new_config: Dict[str, Any] = deepcopy(basis_config)

        # First do replacements among commondef dictionaries themselves
        # Note: These cannot have cycles.
        replacements = self.filter_one_dict(replacements, replacements)

        # Next do replacements among all tools dicts
        tools: List[Dict[str, Any]] = basis_config.get("tools")
        new_config["tools"] = self.filter_one_list(tools, replacements)

        return new_config

    def filter_one_dict(self, source: Dict[str, Any], replacements: Dict[str, Any]) -> Dict[str, Any]:
        """
        Performs the filter on the values of a dictionary

        :param source: The dictionary to filter
        :param replacements: A dictionary of string keys to their replacements
        :return: A new dictionary whose value are (maybe) replaced per the
                 replacements dictionary
        """

        new_dict = deepcopy(source)
        for key, value in source.items():

            replacement_value: Any = None

            if isinstance(value, Dict):
                # We have a dictionary value. Recurse.
                replacement_value = self.filter_one_dict(value, replacements)
            elif isinstance(value, List):
                replacement_value = self.filter_one_list(value, replacements)
            else:
                replacement_value = self.make_replacements(value, replacements)

            if replacement_value is not None:
                new_dict[key] = replacement_value

        return new_dict

    def filter_one_list(self, source: List[Any], replacements: Dict[str, Any]) -> List[Any]:
        """
        Performs the filter on the elements of a list

        :param source: The list to filter
        :param replacements: A dictionary of string keys to their replacements
        :return: A new List whose components are (maybe) replaced per the
                 replacements dictionary
        """

        new_list: List[Any] = []
        for component in source:

            replacement_value: Any = component

            if isinstance(component, Dict):
                replacement_value = self.filter_one_dict(component, replacements)
            elif isinstance(component, List):
                replacement_value = self.filter_one_list(component, replacements)
            else:
                replacement_value = self.make_replacements(component, replacements)

            new_list.append(replacement_value)

        return new_list

    def make_replacements(self, source_value: Any, replacements: Dict[str, Any]) -> Any:
        """
        Make replacements per the keys and values in the replacements dictionary

        :param source_value: The value to potentially do replacements on
        :param replacements: A dictionary of string keys to their replacements
        :return: A potentially new value if some key in the replacements dictionary
                is found to trigger a replacement, otherwise, the same source_value
                that came in.
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""

from typing import Any
from typing import Dict
from typing import List

from copy import deepcopy

from leaf_common.config.config_filter import ConfigFilter
from leaf_common.config.dictionary_overlay import DictionaryOverlay
from leaf_common.parsers.dictionary_extractor import DictionaryExtractor


class DefaultsConfigFilter(ConfigFilter):
    """
    ConfigFilter implementation for copying top-level defaults
    into individual agents.
    """

    # A mapping of source keys for defaults at the top level to destination
    # keys on the specific tool where the top-level defaults (if any) should be copied.
    DEFAULTS_MAPPING: Dict[str, str] = {
        # A value of None implies using the source key as the same destination
        # key in the tool as well.
        "llm_config": None,
        "llm_config.verbose": "verbose",
        "verbose": None,
        "max_iterations": None,
        "max_execution_seconds": None,
        "error_formatter": None,
        "error_fragments": None,
    }

    def filter_config(self, basis_config: Dict[str, Any]) \
            -> Dict[str, Any]:
        """
        Filters the given basis config.

        Ideally this would be a Pure Function in that it would not
        modify the caller's arguments so that the caller has a chance
        to decide whether to take any changes returned.

        :param basis_config: The config dictionary to act as the basis
                for filtering
        :return: A config dictionary, potentially modified as per the
                policy encapsulated by the implementation
        """

        if basis_config is None:
            return basis_config

        tools: List[Dict[str, Any]] = basis_config.get("tools")
        if tools is None or len(tools) == 0:
            # Nothing to do. Exit early.
            return basis_config

        # Do a deepcopy of the basis now that we know we are likely to modify it
        result_config: Dict[str, Any] = deepcopy(basis_config)

        # Create a single extractor for the top-level.
        basis_extractor = DictionaryExtractor(basis_config)
        overlayer = DictionaryOverlay()

        # Loop through all the tools making additions.
        tools = result_config.get("tools")
        for tool in tools:
            tool_extractor = DictionaryExtractor(tool)

            # Loop through all the keys in the defaults mapping.
            for basis_source_key, tool_dest_key in self.DEFAULTS_MAPPING.items():

                basis_value = basis_extractor.get(basis_source_key)
                if basis_value is None:
                    # No value to fill out.
                    continue

                # Account for semantics outlined in default_mapping above
                if tool_dest_key is None:
                    tool_dest_key = basis_source_key

                tool_value = tool_extractor.get(tool_dest_key)
                if tool_value is None:
                    # If the tool does not have a value, use the basis_value whole cloth
                    tool[tool_dest_key] = deepcopy(basis_value)

                elif isinstance(tool_value, Dict) and isinstance(basis_value, Dict):
                    # Do a dictionary overlay with basis_value defaults as what is
                    # overrideable. The values of tool_value are favored.
                    tool[tool_dest_key] = overlayer.overlay(basis_value, tool_value)

                # Otherwise, don't touch the value already in the tool.
                # Don't do anything funny with scalars.
                # Don't try to merge lists (semantics are fragile here).

        return result_config




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from pathlib import Path
import os


class FileOfClass:
    """
    Class which assists in getting absolute paths from a starting source file.
    Clients are expected to invoke with something like:
            FileOfClass(__file__, path_to_basis="../reacharound").
    """

    def __init__(self, source_file: str, path_to_basis: str = "."):
        """
        Constructor

        :param source_file: The source file which will be the starting point for relative paths.
        :param path_to_basis: An optional path to a relative basis, so as to be
                        able to find other relevant files in an absolute manner.
                        This is expected to be a single string of multiple degrees of
                        ".." (or "../..", etc).
                        The default is the directory of the source_file.
        """

        self.source_file: str = source_file
        self.path_to_basis: str = path_to_basis

    def get_file_path(self) -> Path:
        """
        :return: A pathlib Path pointing to the file of the source_file
        """
        return Path(self.source_file)

    def get_file(self) -> str:
        """
        :return: A string pointing to the absolute file path of the source_file
        """
        return str(self.get_file_path().resolve())

    def get_dir_path(self) -> Path:
        """
        :return: A pathlib Path pointing to the directory of the source_file
        """
        return Path(self.source_file).parent

    def get_basis_path(self) -> Path:
        """
        :return: The pathlib Path of the provided basis.
        """
        return self.get_dir_path() / self.path_to_basis

    def get_basis(self) -> str:
        """
        :return: The absolute file path of the provided basis.
        """
        return str(self.get_basis_path().resolve())

    def get_file_in_basis(self, filename: str) -> str:
        """
        :return: An absolute path to a file that resides within the basis.
        """
        return str((self.get_basis_path() / filename).resolve())

    @staticmethod
    def check_file(filepath: str, basis: str = "/") -> str:
        """
        Used to quell Path Traversal vulnerabilities
        :param filepath: The filepath to verify
        :param basis: The directory under which the file must reside
        :return: A fully resolved path if successful.
                Otherwise a ValueError will be thrown.
        """
        if filepath is None:
            return filepath

        if basis is None:
            raise ValueError("basis for file checking is None")

        # Resolve full paths
        # Incorrectly flagged as destination of Path Traversal 1, 2
        #   Reason: This is the method in which we are actually trying to do
        #           the path traversal check itself. CheckMarx does not recognize
        #           pathlib as a valid library with which to resolve these kinds
        #           of issues.
        test_abs_path: str = str(Path(filepath).resolve())

        if basis == "~":
            # Special case if we are looking under user's home directory
            basis = os.path.expanduser("~")
        basis_abs_path: str = str(Path(basis).resolve())

        if not test_abs_path.startswith(basis_abs_path):
            raise ValueError(f"{test_abs_path} must be under {basis_abs_path}")

        return test_abs_path




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Union

from langchain_core.messages.base import BaseMessage


class AgentFrameworkMessage(BaseMessage):
    """
    BaseMessage implementation of a message from the agent framework
    """

    type: Literal["agent-framework"] = "agent-framework"

    def __init__(self, content: Union[str, List[Union[str, Dict]]] = None,
                 chat_context: Dict[str, Any] = None,
                 sly_data: Dict[str, Any] = None,
                 **kwargs: Any) -> None:
        """
        Pass in content as positional arg.

        :param content: The string contents of the message.
        :param chat_context: A dictionary that fully desbribes the state of play
                    of the chat conversation such that when it is passed on to a
                    different server, the conversation can continue uninterrupted.
        :param sly_data: A dictionary of private data, separate from the chat stream.
        :param kwargs: Additional fields to pass to the superclass
        """
        super().__init__(content=content, **kwargs)
        self.chat_context: Dict[str, Any] = chat_context
        self.sly_data: Dict[str, Any] = sly_data




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

import json
import os

from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.internals.messages.origination import Origination
from neuro_san.message_processing.message_processor import MessageProcessor


# pylint: disable=too-many-arguments,too-many-positional-arguments
class ThinkingFileMessageProcessor(MessageProcessor):
    """
    Processes AgentCli input by using the neuro-san streaming API.
    """

    def __init__(self, thinking_file: str, thinking_dir: str):
        """
        Constructor
        """
        self.thinking_file: str = thinking_file
        self.thinking_dir: str = thinking_dir

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """

        # Process any text in the message
        text: str = chat_message_dict.get("text")
        structure: Dict[str, Any] = chat_message_dict.get("structure")
        if text is None and structure is None:
            return

        origin: List[str] = chat_message_dict.get("origin")
        test_origin_str: str = Origination.get_full_name_from_origin(origin)

        origin_str: str = ""
        if test_origin_str is not None:
            origin_str = test_origin_str

        self.write_message(chat_message_dict, origin_str)

    def write_message(self, response: Dict[str, Any], origin_str: str):
        """
        Writes a line of text attributable to the origin, however we are doing that.
        :param response: The message to write
        :param origin_str: The string representing the origin of the message
        """

        response_type: str = response.get("type")
        message_type: ChatMessageType = ChatMessageType.from_response_type(response_type)
        message_type_str: str = ChatMessageType.to_string(message_type)

        text: str = response.get("text")
        structure: Dict[str, Any] = response.get("structure")

        if structure is not None:
            # There is no real text, but there is a structure. JSON-ify it.
            text = json.dumps(structure, indent=4, sort_keys=True)

        filename: str = self.thinking_file
        if self.thinking_dir:
            if origin_str is None or len(origin_str) == 0:
                return
            filename = os.path.join(self.thinking_dir, origin_str)

        how_to_open_file: str = "a"
        if not os.path.exists(filename):
            how_to_open_file = "w"

        with open(filename, how_to_open_file, encoding="utf-8") as thinking:
            use_origin: str = ""

            # Maybe add some context to where message is coming from if not using thinking_dir
            if not self.thinking_dir:
                use_origin += f" from {origin_str}"

            # Maybe add some context as to where the tool result came from if we have info for that.
            tool_result_origin: List[Dict[str, Any]] = response.get("tool_result_origin")
            if tool_result_origin is not None:
                last_origin_only: List[Dict[str, Any]] = [tool_result_origin[-1]]
                origin_str = Origination.get_full_name_from_origin(last_origin_only)
                use_origin += f" (result from {origin_str})"

            # Write the message out
            thinking.write(f"\n[{message_type_str}{use_origin}]:\n")
            thinking.write(text)
            thinking.write("\n")




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List

import logging
import traceback

from openai import BadRequestError

from neuro_san.internals.chat.async_collating_queue import AsyncCollatingQueue
from neuro_san.internals.chat.chat_history_message_processor import ChatHistoryMessageProcessor
from neuro_san.internals.graph.registry.agent_tool_registry import AgentToolRegistry
from neuro_san.internals.graph.tools.front_man import FrontMan
from neuro_san.internals.graph.tools.sly_data_redactor import SlyDataRedactor
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.agent_framework_message import AgentFrameworkMessage
from neuro_san.internals.messages.message_utils import convert_to_chat_message
from neuro_san.internals.run_context.factory.run_context_factory import RunContextFactory
from neuro_san.internals.run_context.interfaces.run_context import RunContext
from neuro_san.message_processing.message_processor import MessageProcessor
from neuro_san.message_processing.answer_message_processor import AnswerMessageProcessor


# pylint: disable=too-many-instance-attributes
class DataDrivenChatSession:
    """
    ChatSession implementation that consolidates policy
    in using data-driven agent tool graphs.
    """

    def __init__(self, registry: AgentToolRegistry):
        """
        Constructor

        :param registry: The AgentToolRegistry to use.
        """
        # This block contains top candidates for state storage that needs to be
        # retained when session_ids go away.
        self.front_man: FrontMan = None

        self.registry: AgentToolRegistry = registry
        self.sly_data: Dict[str, Any] = {}

    async def set_up(self, invocation_context: InvocationContext,
                     chat_context: Dict[str, Any] = None):
        """
        Resets or sets the instance up for the first time.
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        """

        # Reset any sly data
        # This ends up being the one reference to the sly_data that gets passed around
        # to the graph components. Updating this updates everyone else.
        self.sly_data = {}

        # Reset what we might have created before.
        await self.delete_resources()

        run_context: RunContext = RunContextFactory.create_run_context(None, None,
                                                                       invocation_context=invocation_context,
                                                                       chat_context=chat_context)

        self.front_man = self.registry.create_front_man(self.sly_data, run_context)

        await self.front_man.create_resources()

    async def chat(self, user_input: str,
                   invocation_context: InvocationContext,
                   sly_data: Dict[str, Any] = None) -> Iterator[Dict[str, Any]]:
        """
        Main entry-point method for accepting new user input

        :param user_input: A string with the user's input
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be None.
        :return: An Iterator over dictionary representation of chat messages.
                The keys/values/structure of these chat message dictionaries will reflect
                instances of ChatMessage from chat.proto.

                Note that Iterators themselves are *not* simply lists. They are a Python
                construct intended for use in a for-loop that is allowed to come up with
                its content dynamically.  For our purposes, when an initiator of chat()
                gets a handle to this Iterator, they can begin looping/waiting on its contents
                without the content itself having been created yet.  This is a building
                block of streaming results even though direct callers may not actually
                be streaming.
        """
        if self.front_man is None:
            await self.set_up(invocation_context)
        else:
            self.front_man.update_invocation_context(invocation_context)

        # Update sly data, if any.
        # Note that since this instance is the owner of the sly_data,
        # any update here should get transmitted to all the other graph components
        # because it is expected they share the reference and only interact with it
        # in a read-only fashion.
        if sly_data is not None:
            self.sly_data.update(sly_data)

        try:
            # DEF - drill further down for iterator from here to enable getting
            #       messages from downstream agents.
            raw_messages: List[Any] = await self.front_man.submit_message(user_input)

        except BadRequestError:
            # This can happen if the user is trying to send a new message
            # while it is still working on a previous message that has not
            # yet returned.
            raw_messages: List[Any] = [
                AgentFrameworkMessage(content="Patience, please. I'm working on it.")
            ]

            logger = logging.getLogger(self.__class__.__name__)
            logger.error(traceback.format_exc())

        chat_messages: List[Dict[str, Any]] = []
        for raw_message in raw_messages:
            chat_message: Dict[str, Any] = convert_to_chat_message(raw_message, self.front_man.get_origin())
            chat_messages.append(chat_message)

        return iter(chat_messages)

    # pylint: disable=too-many-locals
    async def streaming_chat(self, user_input: str,
                             invocation_context: InvocationContext,
                             sly_data: Dict[str, Any] = None,
                             chat_context: Dict[str, Any] = None):
        """
        Main streaming entry-point method for accepting new user input

        :param user_input: A string with the user's input
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param sly_data: A mapping whose keys might be referenceable by agents, but whose
                 values should not appear in agent chat text. Can be None.
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        :return: Nothing.  Response values are put on a queue whose consumtion is
                managed by the Iterator aspect of AsyncCollatingQueue on the InvocationContext.
        """
        if self.front_man is None:
            await self.set_up(invocation_context, chat_context)

        # Save information about chat
        chat_messages: Iterator[Dict[str, Any]] = await self.chat(user_input, invocation_context, sly_data)
        message_list: List[Dict[str, Any]] = list(chat_messages)

        # Determine the chat_context to enable continuing the conversation
        return_chat_context: Dict[str, Any] = self.prepare_chat_context(message_list)

        # Find "the answer" and have that be the content of the last message we send
        answer_processor = AnswerMessageProcessor()
        answer_processor.process_messages(message_list)
        answer: str = answer_processor.get_answer()
        if answer is None:
            # Can't have content as None or problems arise.
            answer = ""

        # Send back sly_data as the front-man permits
        front_man_spec: Dict[str, Any] = self.front_man.get_agent_tool_spec()
        redactor = SlyDataRedactor(front_man_spec,
                                   config_keys=["allow.to_upstream.sly_data"],
                                   allow_empty_dict=False)
        return_sly_data: Dict[str, Any] = redactor.filter_config(self.sly_data)

        # Stream over chat state as the last message
        message = AgentFrameworkMessage(content=answer, chat_context=return_chat_context,
                                        sly_data=return_sly_data)
        journal: Journal = invocation_context.get_journal()
        await journal.write_message(message, origin=None)

        # Put an end-marker on the queue to tell the consumer we truly are done
        # and it doesn't need to wait for any more messages.
        # The consumer await-s for queue.get()
        queue: AsyncCollatingQueue = invocation_context.get_queue()
        await queue.put_final_item()

    async def delete_resources(self):
        """
        Frees up any service-side resources.
        """
        if self.front_man is not None:
            await self.front_man.delete_resources(None)
            self.front_man = None

    def prepare_chat_context(self, chat_message_history: List[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Prepare the chat context.

        :param chat_message_history: A list of ChatMessage dictionaries that
                comprise the front man's full chat history.
        :return: A ChatContext dictionary comprising the full state of play of
                the conversation such that it could be taken up on a different
                server instance
        """
        # OK if this is None
        max_message_history: int = self.front_man.get_agent_tool_spec().get("max_message_history")
        processor: MessageProcessor = ChatHistoryMessageProcessor(max_message_history)
        processor.process_messages(chat_message_history)

        chat_history: Dict[str, Any] = {
            "origin": self.front_man.get_origin(),
            "messages": processor.get_message_history()
        }

        # For now, we only send the front man's chat history, as that is the
        # state we had been preserving since the early days.  It is conceivable
        # this could expand to other agents in the hierarchy at some point.
        chat_context: Dict[str, Any] = {
            "chat_histories": [chat_history]
        }

        return chat_context




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from neuro_san.internals.run_context.interfaces.agent_spec_provider import AgentSpecProvider
from neuro_san.internals.run_context.interfaces.agent_tool_factory import AgentToolFactory
from neuro_san.internals.run_context.interfaces.run import Run


class ToolCaller(AgentSpecProvider):
    """
    Interface for Tools that call Agents/LLMs as functions.
    This is called by langchain Tools and implemented by CallingTool.
    """

    async def make_tool_function_calls(self, component_run: Run) -> Run:
        """
        Calls all of the callable_components' functions

        :param component_run: The Run which the component is operating under
        :return: A potentially updated Run for the component
        """
        raise NotImplementedError

    def get_factory(self) -> AgentToolFactory:
        """
        :return: The factory that contains the specs of all the tools
        """
        raise NotImplementedError

    def get_agent_tool_spec(self) -> Dict[str, Any]:
        """
        :return: the dictionary describing the data-driven agent
        """
        raise NotImplementedError

    def get_name(self) -> str:
        """
        :return: the name of the data-driven agent as it comes from the spec
        """
        raise NotImplementedError



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
from typing import Any


class AssertForwarder:
    """
    Interface definition for forwarding asserts to different test infrastructure systems.
    """

    # pylint: disable=invalid-name
    def assertEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is equal to the second

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertNotEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is not equal to the second

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertTrue(self, expr: Any, msg: str = None):
        """
        Assert that the expression is true

        :param expr: Expression to test
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertFalse(self, expr: Any, msg: str = None):
        """
        Assert that the expression is false

        :param expr: Expression to test
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertIs(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first and second are the same object

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertIsNot(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first and second are not the same object

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertIsNone(self, expr: Any, msg: str = None):
        """
        Assert that the expression is None

        :param expr: Expression to test
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertIsNotNone(self, expr: Any, msg: str = None):
        """
        Assert that the expression is not None

        :param expr: Expression to test
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertIn(self, member: Any, container: Any, msg: str = None):
        """
        Assert that the member is in the container

        :param member: Member comparison element
        :param container: Container comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertNotIn(self, member: Any, container: Any, msg: str = None):
        """
        Assert that the member is not in the container

        :param member: Member comparison element
        :param container: Container comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertIsInstance(self, obj: Any, cls: Any, msg: str = None):
        """
        Assert that the obj is an instance of the cls

        :param obj: object instance comparison element
        :param cls: Class comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertNotIsInstance(self, obj: Any, cls: Any, msg: str = None):
        """
        Assert that the obj is not an instance of the cls

        :param obj: object instance comparison element
        :param cls: Class comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertGreater(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is greater than the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertGreaterEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is greater than or equal to the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertLess(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is less than the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertLessEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is less than or equal to the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertGist(self, gist: bool, acceptance_criteria: str, text_sample: str, msg: str = None):
        """
        Assert that the gist is true

        :param gist: Pass/Fail value of the gist expected to be True
        :param acceptance_criteria: The value to verify against
        :param text_sample: The value appearing in the test sample
        :param msg: optional string message
        """
        raise NotImplementedError

    # pylint: disable=invalid-name
    def assertNotGist(self, gist: bool, acceptance_criteria: str, text_sample: str, msg: str = None):
        """
        Assert that the gist is true

        :param gist: Pass/Fail value of the gist expected to be False
        :param acceptance_criteria: The value to verify against
        :param text_sample: The value appearing in the test sample
        :param msg: optional string message
        """
        raise NotImplementedError



# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from typing import Dict
from typing import List

import logging
import os

from leaf_server_common.logging.logging_setup import setup_logging
from leaf_server_common.server.grpc_metadata_forwarder import GrpcMetadataForwarder

from neuro_san.internals.utils.file_of_class import FileOfClass


class AgentServerLogging:
    """
    Common logging setup for the agent server threads.
    """

    def __init__(self, server_name_for_logs: str,
                 forwarded_request_metadata_str: str):
        """
        Constructor

        :param server_name_for_logs: The server name (source) to be used in structured logging messages
        :param forwarded_request_metadata_str: A space-delimited string of http header metadata
                            whose key/value pairs are to be forwarded on to the logging system.
                            Note that individual keys must be snake_case. No capitals.
                            (I guess per HTTP rules).
        """
        self.server_name_for_logs: str = server_name_for_logs
        self.forwarded_request_metadata: List[str] = forwarded_request_metadata_str.split(" ")

    def get_forwarder(self) -> GrpcMetadataForwarder:
        """
        :return: A GrpcMetadataForwarder instance initialized with
                 the list of forwarded request metadata keys
        """
        return GrpcMetadataForwarder(self.forwarded_request_metadata)

    def setup_logging(self, metadata: Dict[str, str] = None, request_id: str = "None"):
        """
        Set up logging for agent server threads.

        :param metadata: An optional dictionary with actual metadata
        :param request_id: An optional request_id string.  Default is "None".
        """

        # Make for easy running from the neuro-san repo
        if os.environ.get("AGENT_SERVICE_LOG_JSON") is None:
            # Use the log file that is local to the repo
            file_of_class = FileOfClass(__file__, path_to_basis="../deploy")
            os.environ["AGENT_SERVICE_LOG_JSON"] = file_of_class.get_file_in_basis("logging.json")

        # Need to initialize the forwarded metadata default values before our first
        # call to a logger (which is below!).
        extra_logging_defaults: Dict[str, str] = {
            "source": self.server_name_for_logs,
            "user_id": "None",
            "request_id": request_id,
        }
        if len(self.forwarded_request_metadata) > 0:
            for key in self.forwarded_request_metadata:
                if metadata is not None:
                    extra_logging_defaults[key] = metadata.get(key, "None")
                else:
                    extra_logging_defaults[key] = "None"

        current_dir: str = os.path.dirname(os.path.abspath(__file__))
        setup_logging(self.server_name_for_logs, current_dir,
                      'AGENT_SERVICE_LOG_JSON',
                      'AGENT_SERVICE_LOG_LEVEL',
                      extra_logging_defaults)

        # This module within openai library can be quite chatty w/rt http requests
        logging.getLogger("httpx").setLevel(logging.WARNING)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.internals.interfaces.agent_tool_factory_provider import AgentToolFactoryProvider


class ToolFactoryProvider:
    """
    Abstract interface for getting agent-specific tool factory providers
    """

    def get_agent_tool_factory_provider(self, agent_name: str) -> AgentToolFactoryProvider:
        """
        Get tool factory provider for given agent.
        :param agent_name: name of an agent
        :return: tool factory provider for this agent,
                 or None if no such provider is available.
        """
        raise NotImplementedError




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any, Dict, Generator
import json

from google.protobuf.json_format import MessageToDict
from google.protobuf.json_format import Parse

# pylint: disable=no-name-in-module
from neuro_san.api.grpc.agent_pb2 import ChatRequest, ChatResponse

from neuro_san.http_sidecar.handlers.base_request_handler import BaseRequestHandler
from neuro_san.interfaces.async_agent_session import AsyncAgentSession


class StreamingChatHandler(BaseRequestHandler):
    """
    Handler class for neuro-san streaming chat API call.
    """

    async def stream_out(self,
                         generator: Generator[Generator[ChatResponse, None, None], None, None]) -> int:
        """
        Process streaming out generator output to HTTP connection.
        :param generator: async gRPC generator
        :return: number of chat responses streamed out.
        """
        # Set up headers for chunked response
        self.set_header("Content-Type", "application/json-lines")
        self.set_header("Transfer-Encoding", "chunked")
        # Flush headers immediately
        flush_ok: bool = await self.do_flush()
        if not flush_ok:
            return 0

        sent_out: int = 0
        async for sub_generator in generator:
            async for result_message in sub_generator:
                result_dict: Dict[str, Any] = MessageToDict(result_message)
                result_str: str = json.dumps(result_dict) + "\n"
                self.write(result_str)
                flush_ok = await self.do_flush()
                if not flush_ok:
                    return sent_out
                sent_out += 1
        return sent_out

    async def post(self, agent_name: str):
        """
        Implementation of POST request handler for streaming chat API call.
        """

        metadata: Dict[str, Any] = self.get_metadata()
        update_done: bool = await self.update_agents(metadata)
        if not update_done:
            return

        if not self.agent_policy.allow(agent_name):
            self.set_status(404)
            self.logger.error({}, "error: Invalid request path %s", self.request.path)
            self.do_finish()
            return

        self.logger.info(metadata, "Start POST %s/streaming_chat", agent_name)
        sent_out = 0
        try:
            # Parse JSON body
            data = json.loads(self.request.body)

            grpc_request = Parse(json.dumps(data), ChatRequest())
            grpc_session: AsyncAgentSession = self.get_agent_grpc_session(metadata, agent_name)

            # Mind the type hint:
            # here we are getting Generator of Generators of ChatResponses!
            result_generator: Generator[Generator[ChatResponse, None, None], None, None] =\
                grpc_session.streaming_chat(grpc_request)
            sent_out = await self.stream_out(result_generator)

        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.process_exception(exc)
        finally:
            # We are done with response stream:
            self.do_finish()
            self.logger.info(metadata, "Finish POST %s/streaming_chat %d responses", agent_name, sent_out)




# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

import asyncio
from typing import Any
from typing import Dict
from typing import List
from typing_extensions import override

from langchain_community.callbacks.bedrock_anthropic_callback import MODEL_COST_PER_1K_INPUT_TOKENS
from langchain_community.callbacks.bedrock_anthropic_callback import MODEL_COST_PER_1K_OUTPUT_TOKENS
from langchain_core.callbacks import AsyncCallbackHandler
from langchain_core.messages import AIMessage
from langchain_core.messages.ai import UsageMetadata
from langchain_core.outputs import ChatGeneration, LLMResult

EMPTY = ""


def calculate_anthropic_token_cost(input_tokens: int, output_tokens: int, model_name: str) -> float:
    """
    Calculate the token cost for an Anthropic Claude model from lookup table MODEL_COST_PER_1K_INPUT_TOKENS
    and MODEL_COST_PER_1K_OUTPUT_TOKENS.

    This function allows users to input a partial model name (e.g., 'claude-3-7-sonnet-20250219')
    instead of requiring the full internal model ID (e.g., 'anthropic.claude-3-7-sonnet-20250219-v1:0').

    Since users often do not know or use the full system-defined model names,
    we perform a flexible substring match across all known models:
    - If exactly one match is found, we calculate the cost based on that model.
    - If multiple matches are found, we return 0.0 and print an error to alert the user.
    - If no matches are found, we also return 0.0 and print an error.

    :param prompt_tokens: Number of input (prompt) tokens.
    :param completion_tokens: Number of output (completion) tokens.
    :param model_name: A model name (e.g., 'claude-3-7-sonnet-20250219').

    :return: The total cost as a float.
    """

    if not model_name:
        print("[Error] model_name must be provided.")
        return 0.0

    # Check if "model_name" is in any keys of the lookup table.
    matching_models: List[str] = [
        model for model in MODEL_COST_PER_1K_INPUT_TOKENS if model_name in model
    ]

    # Return cost = 0.0 if there are no match or more than one matches.
    if not matching_models:
        print(
            f"[Error] Unknown model: '{model_name}'. No matches found. "
            "Known models are: " + ", ".join(MODEL_COST_PER_1K_INPUT_TOKENS.keys())
        )
        return 0.0

    if len(matching_models) > 1:
        print(
            f"[Error] Ambiguous model name '{model_name}'. Matches multiple models: "
            + ", ".join(matching_models)
        )
        return 0.0

    full_model_id: str = matching_models[0]

    input_cost: float = (input_tokens / 1000) * MODEL_COST_PER_1K_INPUT_TOKENS[full_model_id]
    output_cost: float = (output_tokens / 1000) * MODEL_COST_PER_1K_OUTPUT_TOKENS[full_model_id]
    return input_cost + output_cost


# pylint: disable=too-many-ancestors
class LlmTokenCallbackHandler(AsyncCallbackHandler):
    """
    Callback handler that tracks token usage via "AIMessage.usage_metadata".

    This class is a modification of LangChain’s "UsageMetadataCallbackHandler" and "OpenAICallbackHandler":
    - https://python.langchain.com/api_reference/_modules/langchain_core/callbacks/usage.html
    #get_usage_metadata_callback
    - https://python.langchain.com/api_reference/_modules/langchain_community/callbacks/openai_info.html
    #OpenAICallbackHandler

    It collects token usage from the "usage_metadata" field of "AIMessage" each time an LLM or chat model
    finishes execution.
    The metadata is a dictionary that may include:
    - "input_tokens" (collected as "prompt_tokens")
    - "output_tokens" (collected as "completion_tokens")
    - "total_tokens"

    This handler tracks these values internally and is compatible with models that populate "usage_metadata",
    regardless of provider.

    **Note**: While the "AIMessage.response_data" may contain model names, they are not currently included in the
    report format. A future version may support returning usage statistics grouped by model name.

    Example of expected future output structure (not currently implemented):

        {
            "gpt-4o-mini-2024-07-18": {
                "input_tokens": 8,
                "output_tokens": 10,
                "total_tokens": 18,
                "input_token_details": {"audio": 0, "cache_read": 0},
                "output_token_details": {"audio": 0, "reasoning": 0}
            },
            "claude-3-5-haiku-20241022": {
                "input_tokens": 8,
                "output_tokens": 21,
                "total_tokens": 29,
                "input_token_details": {"cache_read": 0, "cache_creation": 0}
            }
        }

    Note:
    This class is intended for use with Ollama models, as OpenAICallbackHandler and
    BedrockAnthropicTokenUsageCallbackHandler already handle OpenAI and Anthropic models, respectively.
    Ollama models currently have no associated cost, so total_cost is always set to 0.0 to maintain compatibility
    with reporting templates.
    """

    total_tokens: int = 0
    prompt_tokens: int = 0
    completion_tokens: int = 0
    successful_requests: int = 0
    total_cost: float = 0.0

    def __init__(self) -> None:
        """Initialize the CallbackHandler."""
        super().__init__()
        self._lock = asyncio.Lock()

    @override
    def __repr__(self) -> str:
        return (
            f"Tokens Used: {self.total_tokens}\n"
            f"\tPrompt Tokens: {self.prompt_tokens}\n"
            f"\tCompletion Tokens: {self.completion_tokens}\n"
            f"Successful Requests: {self.successful_requests}\n"
            f"Total Cost (USD): ${self.total_cost}"
        )

    @override
    async def on_llm_end(self, response: LLMResult, **kwargs: Any) -> None:
        """
        Collect token usage when llm ends.
        :param response: Output from chat model
        """
        # Check for usage_metadata (Only work for langchain-core >= 0.2.2)
        try:
            generation = response.generations[0][0]
        except IndexError:
            generation = None

        usage_metadata: UsageMetadata = None
        response_metadata: Dict[str, Any] = None
        model_name: str = EMPTY
        if isinstance(generation, ChatGeneration):
            try:
                message = generation.message
                if isinstance(message, AIMessage):
                    # Token info is in an attribute of AIMessage called "usage_metadata".
                    usage_metadata = message.usage_metadata
                    # Get model name so that cost can be determined if needed.
                    response_metadata = message.response_metadata
                    if response_metadata:
                        if "model_name" in response_metadata:
                            model_name = response_metadata.get("model_name")
                        elif "model_id" in response_metadata:
                            model_name = response_metadata.get("model_id")
                        elif "model" in response_metadata:
                            model_name = response_metadata.get("model")
            except AttributeError:
                pass

        if usage_metadata:
            total_tokens: int = usage_metadata.get("total_tokens", 0)
            completion_tokens: int = usage_metadata.get("output_tokens", 0)
            prompt_tokens: int = usage_metadata.get("input_tokens", 0)

            # If it is an anthropic model, calculate the cost.
            if "claude" in model_name:
                total_cost: float = calculate_anthropic_token_cost(prompt_tokens, completion_tokens, model_name)
            else:
                total_cost = 0.0

            # update shared state behind lock
            async with self._lock:
                self.total_tokens += total_tokens
                self.prompt_tokens += prompt_tokens
                self.completion_tokens += completion_tokens
                self.successful_requests += 1
                self.total_cost += total_cost
